import base64
import json
import pathlib

import pytest
from timbal.types import File, FileContent, Message, TextContent, ToolResultContent, ToolUseContent


def test_message_text_validation() -> None:
    message = Message(role="assistant", content=[TextContent(text="Hello, World!")])
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert len(message.content) == 1
    assert message.content[0].type == "text"
    assert message.content[0].text == "Hello, World!"
    assert message.content == [TextContent(text="Hello, World!")]

    # text must be a string
    with pytest.raises(ValueError):
        Message.validate({"role": "assistant", "content": [{"type": "text", "text": 123}]})


def test_message_text_to_openai_input() -> None:
    message = Message(role="assistant", content=[TextContent(text="Hello, World!")])
    assert message.to_openai_input() == {"role": "assistant", "content": [{"type": "text", "text": "Hello, World!"}]}


def test_message_from_openai_input() -> None:
    message = Message.validate({"role": "assistant", "content": [{"type": "text", "text": "Hello, World!"}]})
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert message.content == [TextContent(text="Hello, World!")]


def test_message_text_to_anthropic_input() -> None:
    message = Message(role="assistant", content=[TextContent(text="Hello, World!")])
    assert message.to_anthropic_input() == {"role": "assistant", "content": [{"type": "text", "text": "Hello, World!"}]}


def test_message_from_anthropic_input() -> None:
    message = Message.validate({"role": "assistant", "content": [{"type": "text", "text": "Hello, World!"}]})
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert message.content == [TextContent(text="Hello, World!")]


def test_message_file_validation(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content)
    file_content = FileContent(file=File.validate(str(test_file)))
    message = Message(role="assistant", content=[file_content])
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert isinstance(message.content[0], FileContent)

    # image_url must be a File
    with pytest.raises(ValueError):
        Message.validate({"role": "assistant", "content": [{"type": "image_url", "image_url": {"url": "not a file"}}]})
    

def test_message_file_to_openai_input(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content)

    encoded_image = base64.b64encode(png_content).decode("utf-8")

    message = Message(role="assistant", content=[FileContent(file=File.validate(str(test_file)))])
    assert message.to_openai_input() == {
        "role": "assistant", 
        "content": [{
            "type": "image_url", 
            "image_url": {"url": f"data:image/png;base64,{encoded_image}"}
        }]
    }


def test_message_file_to_anthropic_input(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content)
    message = Message(role="assistant", content=[FileContent(file=File.validate(str(test_file)))])
    assert message.to_anthropic_input() == {"role": "assistant", "content": [{"type": "image", "source": {"type": "base64", "media_type": "image/png", "data": base64.b64encode(png_content).decode("utf-8")}}]}


def test_message_tool_use_validation() -> None:
    message = Message(role="assistant", content=[ToolUseContent(id="123", name="get_weather", input={"city": "London"})])
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert message.content == [ToolUseContent(id="123", name="get_weather", input={"city": "London"})]

    # input must be a dict
    with pytest.raises(ValueError):
        Message.validate({"role": "assistant", "content": [{"type": "tool_use", "id": "123", "name": "get_weather", "input": "not a dict"}]})


def test_message_with_tool_use_to_openai_input() -> None:
    message = Message(role="assistant", content=[ToolUseContent(id="123", name="get_weather", input={"city": "London"})])
    assert message.to_openai_input() == {"role": "assistant", "tool_calls": [{"id": "123", "type": "function", "function": {"arguments": '{"city": "London"}', "name": "get_weather"}}]}


def test_message_with_tool_use_from_openai_input() -> None:
    message = Message.validate({"role": "assistant", "tool_calls": [{"id": "123", "type": "function", "function": {"arguments": '{"city": "London"}', "name": "get_weather"}}]})
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert message.content == [ToolUseContent(id="123", name="get_weather", input={"city": "London"})]


def test_message_with_tool_use_to_anthropic_input() -> None:
    message = Message(role="user", content=[ToolUseContent(id="123", name="get_weather", input={"city": "London"})])
    assert message.to_anthropic_input() == {"role": "user", "content": [{"type": "tool_use", "id": "123", "name": "get_weather", "input": {"city": "London"}}]}


def test_message_with_tool_result_from_anthropic_input() -> None:
    message = Message.validate({"role": "user", "content": [{"type": "tool_result", "tool_use_id": "123", "content": [{"type": "text", "text": "Hello, World!"}]}]})
    assert isinstance(message, Message)
    assert message.role == "user"
    assert message.content == [ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])]


def test_message_tool_result_validation() -> None:
    message = Message(role="assistant", content=[ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])])
    assert isinstance(message, Message)
    assert message.role == "assistant"
    assert message.content == [ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])]

    # content must be a list
    with pytest.raises(ValueError):
        Message.validate({"role": "assistant", "content": [{"type": "tool_result", "tool_use_id": "123", "content": 123}]})


def test_message_with_tool_result_to_openai_input() -> None:
    message = Message(role="user", content=[ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])])
    assert message.to_openai_input() == {"role": "tool", "tool_call_id": "123", "content": [{"type": "text", "text": "Hello, World!"}]}


def test_message_with_tool_result_from_openai_input() -> None:
    message = Message.validate({"role": "tool", "tool_call_id": "123", "content": json.dumps({
        "city": "New York",
        "weather": "cloudy"
    })})
    assert isinstance(message, Message)
    assert message.role == "user"
    assert message.content == [ToolResultContent(id="123", content=[TextContent(text='{"city": "New York", "weather": "cloudy"}')])]

    message = Message.validate({"role": "tool", "tool_call_id": "123", "content": [{"type": "text", "text": '{"city": "New York", "weather": "cloudy"}'}]})
    assert isinstance(message, Message)
    assert message.role == "user"
    assert message.content == [ToolResultContent(id="123", content=[TextContent(text='{"city": "New York", "weather": "cloudy"}')])]


def test_message_with_tool_result_to_anthropic_input() -> None:
    message = Message(role="user", content=[ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])])
    assert message.to_anthropic_input() == {"role": "user", "content": [{"type": "tool_result", "tool_use_id": "123", "content": [{"type": "text", "text": "Hello, World!"}]}]}



from typing import Any

import pytest
from timbal import Flow
from timbal.types import Field


def identity_handler(x: Any = Field(default=None)) -> Any:
    return x


@pytest.mark.asyncio
async def test_string_interpolation():
    flow = (
        Flow()
        .add_step("step_1", identity_handler)
        .add_step("step_2", identity_handler)
        .add_step("step_3", identity_handler)
        .add_link("step_1", "step_3")
        .add_link("step_2", "step_3")
        .set_data_map("step_1.x", "x1")
        .set_data_map("step_2.x", "x2")
        .set_data_value("step_3.x",  "{{step_1.return}} {{step_2.return}}")
        .set_output("step_3.x", "step_3_x")
        .set_output("step_3.return", "step_3_return")
    )

    x1 = "Hello"
    x2 = "World"

    async for event in flow.run(x1=x1, x2=x2):
        if event.type == "FLOW_OUTPUT":
            flow_output = event.output
            assert flow_output["step_3_x"] == f"{x1} {x2}"
            assert flow_output["step_3_return"] == f"{x1} {x2}"



from datetime import datetime

import pytest
from timbal import Agent
from timbal.state.savers import InMemorySaver
from timbal.types import Field, Message


def get_datetime() -> str:
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def get_weather(
    location: str = Field(description="The location to get the weather for"), # noqa: ARG001
) -> str:
    return "15 celsius"


def get_traffic(
    location: str = Field(description="The location to get the weather for"), # noqa: ARG001
) -> str:
    return "pretty heavy"


def convert_to_fahrenheit(
    temperature: str = Field(description="The temperature to convert to fahrenheit"), # noqa: ARG001
) -> str:
    return "59 fahrenheit"


@pytest.mark.asyncio
async def test_agent():
    prompt="What's the weather and traffic like in New York?"

    flow = Agent(tools=[get_weather, get_traffic])

    flow_output_event = await flow.complete(prompt=prompt)

    assert isinstance(flow_output_event.output, Message)

    message = flow_output_event.output
    assert "15" in message.content[0].text
    assert "pretty heavy" in message.content[0].text.lower()
    

@pytest.mark.asyncio
async def test_agent_with_tool_description():
    prompt="What's the weather and traffic like in New York?"

    flow = Agent(
        model="gpt-4o-mini", 
        tools=[
            get_weather, 
            {
                "tool": get_traffic, 
                "description": "Get the traffic for a given location"
            }
        ],
    ).compile(state_saver=InMemorySaver())

    flow_output_event = await flow.complete(prompt=prompt)

    assert isinstance(flow_output_event.output, Message)

    message = flow_output_event.output
    assert "15" in message.content[0].text
    assert "pretty heavy" in message.content[0].text.lower()


@pytest.mark.asyncio
async def test_agent_with_multiple_iterations():
    prompt = "What's the weather in New York in fahrenheit?"

    flow = Agent(
        model="gpt-4o-mini", 
        tools=[get_weather, get_traffic, convert_to_fahrenheit],
        max_iter=2
    ).compile(state_saver=InMemorySaver())

    flow_output_event = await flow.complete(prompt=prompt)

    assert isinstance(flow_output_event.output, Message)

    message = flow_output_event.output
    assert "59" in message.content[0].text.lower()
    assert "pretty heavy" not in message.content[0].text.lower()


@pytest.mark.asyncio
async def test_agent_with_no_tools():
    prompt = "Hello are you an LLM with no tools?"

    flow = Agent()

    flow_output_event = await flow.complete(prompt=prompt)

    assert isinstance(flow_output_event.output, Message)


@pytest.mark.asyncio
async def test_agent_with_custom_outputs():
    flow = Agent(tools=[get_datetime])
    flow.set_output("agent_get_datetime_0.return", "datetime")

    assert not isinstance(flow.return_model(), Message)

    prompt = "What's the current time?"
    flow_output_event = await flow.complete(prompt=prompt)
    assert "datetime" in flow_output_event.output
    assert flow_output_event.output["datetime"] is not None

    prompt = "What are you?"
    flow_output_event = await flow.complete(prompt=prompt)
    assert "datetime" in flow_output_event.output
    assert flow_output_event.output["datetime"] is None



from typing import Any

import pytest
from timbal import Flow
from timbal.errors import InvalidLinkError
from timbal.types import Field


def identity_handler(x: Any = Field(default=None)) -> Any:
    return x


@pytest.fixture
def flow():
    return (
        Flow(id="test_add_link")
        .add_step("step_1", identity_handler)
        .add_llm("step_2")
        .add_step("step_3", identity_handler)
    )


def test_add_link(flow: Flow):
    flow.add_link("step_1", "step_2")
    assert "step_1-step_2" in flow.links

    # Test invalid steps ids
    with pytest.raises(InvalidLinkError):
        flow.add_link("step_1", "step_4")
    with pytest.raises(InvalidLinkError):
        flow.add_link("step_4", "step_2")


def test_already_linked(flow: Flow):
    flow.add_link("step_1", "step_2")
    flow.add_link("step_2", "step_3")

    with pytest.raises(InvalidLinkError):
        flow.add_link("step_1", "step_3")


def test_cycle(flow: Flow):
    flow.add_link("step_1", "step_2")
    flow.add_link("step_2", "step_3")

    with pytest.raises(InvalidLinkError):
        flow.add_link("step_3", "step_1")

    with pytest.raises(InvalidLinkError):
        flow.add_link("step_3", "step_3")


def test_tool(flow: Flow):
    with pytest.raises(InvalidLinkError):
        flow.add_link("step_1", "step_3", is_tool=True)

    flow.add_link("step_2", "step_3", is_tool=True)
    step_2_tools = flow.data["step_2.tools"].resolve(context_data=None)
    assert all(tool.name == "step_3" for tool in step_2_tools)


def test_tool_result(flow: Flow):
    flow.add_link("step_1", "step_2", is_tool_result=True)

    with pytest.raises(InvalidLinkError):
        flow.add_link("step_2", "step_3", is_tool_result=True)



from collections.abc import Callable
from typing import Any

from ..errors import DataKeyError, FlowExecutionError
from ..state.data import BaseData, DataError, get_data_key
from ..types.message import Message
from .base import BaseStep
from .flow import Flow
from .utils import get_sources


class Agent(Flow):
    """Subclass of Flow that implements an LLM agent with tool-use capabilities.
    
    An Agent is a specialized Flow that creates a chain of LLM steps and tools, allowing
    the LLM to use tools multiple times in a conversation. Each LLM step can call any
    available tool, and the results are fed back to the next LLM step.

    Attributes:
        id (str): Unique identifier for the agent, defaults to "agent"
        tools (list): List of callable functions, BaseSteps, or dicts with tool configs
        max_iter (int): Maximum number of tool use iterations allowed, defaults to 1
        **kwargs: Additional keyword arguments for the LLMs

    Example:
        ```python
        agent = Agent(
            tools=[search_tool, calculator],
            max_iter=3,
            system_prompt="You are a helpful assistant"
        )
        ```
    """

    def __init__(
        self,
        id: str | None = None,
        tools: list[Callable | BaseStep | dict[str, Any]] = [],
        max_iter: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize an Agent instance.

        Args:
            id: Unique identifier for the agent, defaults to "agent"
            tools: List of tools the agent can use
            max_iter: Maximum number of tool use iterations
            **kwargs: Additional LLM configuration options
        """
        if id is None:
            id = "agent"

        # We don't allow to pass a memory_id here. Agents will always use the same memory.
        # And they must have one, so they can use tool results with the corresponding tool calls.
        kwargs.pop("memory_id", None)
        memory_id = id

        kwargs_system_prompt = kwargs.pop("system_prompt", None)

        super().__init__(id=id)

        # Create initial LLM step that receives the prompt
        entrypoint_llm_id = f"{id}_llm_0"
        self.add_llm(
            id=entrypoint_llm_id,
            memory_id=memory_id,
            system_prompt=kwargs_system_prompt,
            **kwargs,
        )

        # This is the default input for the agent. We don't expose any other params as of yet.
        # If the user wants to pass other inputs, they can do so by calling set_input() on the agent instance afterwards.
        self.set_input(f"{entrypoint_llm_id}.prompt", "prompt")

        if not len(tools):
            return

        # Default system prompt when processing tool results. Give the ability to the user to override this from the outside.
        tool_result_llm_system_prompt = """Please process the tool result given the call. 
        Do not remove references or citations from text. Leave them in markdown format."""
        if kwargs_system_prompt is not None:
            tool_result_llm_system_prompt = kwargs_system_prompt

        prev_llm_id = entrypoint_llm_id
        # Create additional LLM steps for each iteration, allowing the agent to use tools multiple times
        for i in range(max_iter):
            iter_llm_id = f"{id}_llm_{i + 1}"
            self.add_llm(
                id=iter_llm_id,
                memory_id=memory_id,
                system_prompt=tool_result_llm_system_prompt,
                **kwargs,
            )

            # For each tool, create a step and add links for tool use and tool result
            for tool_item in tools:
                if isinstance(tool_item, dict):
                    tool = tool_item["tool"]
                    description = tool_item.get("description")
                else:
                    tool = tool_item
                    description = None

                if callable(tool):
                    tool_id = tool.__name__
                elif isinstance(tool, BaseStep):
                    tool_id = tool.id

                iter_tool_id = f"{id}_{tool_id}_{i}"

                if iter_tool_id not in self.steps:
                    self.add_step(iter_tool_id, tool)

                self.add_link(
                    step_id=prev_llm_id,
                    description=description,
                    next_step_id=iter_tool_id,
                    is_tool=True,
                )

                self.add_link(
                    step_id=iter_tool_id,
                    next_step_id=iter_llm_id,
                    is_tool_result=True,
                )
    
            prev_llm_id = iter_llm_id


    def _collect_outputs(self, data: dict[str, BaseData]) -> Any:
        """Override the method to collect the last available LLM output from the flow always."""
        if len(self.outputs):
            return super()._collect_outputs(data)

        _, rev_dag = self.get_dags()
        rev_sources = list(get_sources(rev_dag))
        assert len(rev_sources) == 1, "Tool to LLM agent mode should have a single LLM as last step"

        last_llm_output = None
        last_llm_id = rev_sources[0]
        while last_llm_output is None:
            try: 
                last_llm_output_key = f"{last_llm_id}.return"
                last_llm_output = get_data_key(data, last_llm_output_key)
                if isinstance(last_llm_output, DataError):
                    flow_execution_error_key = f"{self.path}.{last_llm_id}"
                    raise FlowExecutionError(f"Error collecting outputs {{'{flow_execution_error_key}'}}.")
            except DataKeyError:
                last_llm_output = None

                last_tools_ids = list(rev_dag[last_llm_id])
                assert len(last_tools_ids) >= 1, \
                    "The last LLM of a tool to LLM agent must have at least one tool result."
                last_tools_errors = set()
                for last_tool_id in last_tools_ids:
                    last_tool_output_key = f"{last_tool_id}.return"
                    try:
                        last_tool_output = get_data_key(data, last_tool_output_key)
                        if isinstance(last_tool_output, DataError):
                            flow_execution_error_key = f"{self.path}.{last_tool_id}"
                            last_tools_errors.add(flow_execution_error_key)
                    except DataKeyError:
                        pass

                last_llms_ids = list(rev_dag[last_tools_ids[0]])
                assert len(last_llms_ids) == 1, \
                    "Tool to LLM agent mode should have a single LLM as last step"
                last_llm_id = last_llms_ids[0]
        return last_llm_output

    
    def return_model(self) -> Any:
        """Override the return model to return always a Message."""
        if len(self.outputs):
            return super().return_model()

        return Message



from typing import Any

from anthropic.types import (
    Message as AnthropicMessage,
)
from openai.types.chat import (
    ChatCompletionMessage as OpenAIMessage,
)
from pydantic import (
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    SerializationInfo,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
)
from pydantic_core import CoreSchema, core_schema

from .chat.content import Content, ToolResultContent, ToolUseContent


class Message:
    """A class representing a message in a conversation with an LLM.
    
    This class handles messages for both OpenAI and Anthropic formats, providing
    conversion methods between different message formats and validation logic.
    
    Attributes:
        role: The role of the message sender
        content: The content of the message, which can include text and tool interactions
    """

    __slots__ = ("role", "content")


    def __init__(
        self,
        role: Any,
        content: Any,
    ) -> None:
        """Initialize a Message instance.
        
        Args:
            role: The role of the message sender
            content: The content of the message
        """
        object.__setattr__(self, "role", role)
        object.__setattr__(self, "content", content)


    def __str__(self) -> str:
        return f"Message(role={self.role}, content={self.content})"


    def __repr__(self) -> str:
        return f"Message(role={self.role}, content={self.content})"


    def to_openai_input(self) -> dict[str, Any]:
        """Convert the message to OpenAI's expected input format."""
        # OpenAI expects tool calls to be in a separate field in the message
        role = self.role
        content = []
        tool_calls = []
        for content_item in self.content:
            if isinstance(content_item, ToolUseContent):
                tool_calls.append(content_item.to_openai_input())
            elif isinstance(content_item, ToolResultContent):
                return content_item.to_openai_input()
            else:
                content.append(content_item.to_openai_input())
        openai_input = {"role": role,}
        if len(content):
            openai_input["content"] = content 
        if len(tool_calls): 
            openai_input["tool_calls"] = tool_calls

        return openai_input
    

    def to_anthropic_input(self) -> dict[str, Any]:
        """Convert the message to Anthropic's expected input format."""
        return {
            "role": self.role,
            "content": [item.to_anthropic_input() for item in self.content],
        }


    @classmethod
    def validate(cls, value: ValidatorFunctionWrapHandler, _info: dict | ValidationInfo | None = None) -> "Message":
        """Validate and convert inputs into a Message instance."""
        # Don't recurse if we're already dealing with a Message instance
        if isinstance(value, Message):
            return value

        elif isinstance(value, OpenAIMessage):
            role = value.role
            tool_calls = value.tool_calls
            # If there are tool calls in the message, for us it will be ToolUseContent
            if tool_calls:
                content = [Content.model_validate(item) for item in tool_calls]
                return cls(role=role, content=content)
            content = value.content
            if not isinstance(content, list):
                content = [content]
            content = [Content.model_validate(item) for item in content]
            return cls(role=role, content=content)
 
        elif isinstance(value, AnthropicMessage):
            role = value.role
            content = value.content
            if not isinstance(content, list):
                raise ValueError(f"Invalid content: {content}")
            content = [Content.model_validate(item) for item in content]
            return cls(role=role, content=content)

        elif isinstance(value, dict):
            role = value.get("role", None)
            content = value.get("content", None)
            tool_calls = value.get("tool_calls", None)
            # If there are tool calls in the message, for us it will be ToolUseContent
            if tool_calls:
                content = [Content.model_validate(item) for item in tool_calls]
                return cls(role=role, content=content)
            # If the role is "tool", for us it will be a ToolResultContent
            if role == "tool":
                if isinstance(content, list):
                    content_list = content
                else:
                    content_list = [content]
                content = ToolResultContent(
                    id=value.get("tool_call_id"), 
                    content=[Content.model_validate(item) for item in content_list],
                )
                return cls(role="user", content=[content])
            if not isinstance(content, list):
                content = [content]
            content = [Content.model_validate(item) for item in content]
            return cls(role=role, content=content)
        
        return cls.validate({
            "role": "user",
            "content": value,
        })


    @classmethod
    def serialize(cls, value: Any, _info: dict | SerializationInfo | None = None) -> str:
        """Serialize a Message instance into a dictionary format."""
        # When creating a model with fields with File type that are nullable,
        # pydantic will pass None as the value to File.serialize.
        if value is None:
            return None

        if not isinstance(value, cls):
            raise ValueError("Cannot serialize a non-message object.")

        return {
            "role": value.role,
            "content": value.content,
        }


    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema: CoreSchema, _handler: GetJsonSchemaHandler) -> dict[str, Any]:
        """Defines what this type should be in openapi.json."""
        # https://docs.pydantic.dev/2.8/errors/usage_errors/#custom-json-schema
        json_schema = {
            "title": "TimbalMessage",  # This becomes the type name in most generators
            "type": "object",
            "properties": {
                "role": {
                    "type": "string",
                    "enum": ["user", "assistant", "system"],
                },
                "content": {
                    "type": "array",
                    "items": {}, # Keep it open/generic for now.
                }
            },
        }
        return json_schema


    @classmethod
    def __get_pydantic_core_schema__(cls, _source: type[Any], _handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Defines how to serialize this type in the core schema."""
        return core_schema.with_info_plain_validator_function(
            cls.validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.serialize,
                info_arg=True,
                when_used="always",
            ),
        )



from typing import Any

from ...types import Field, Message, Tool
from .anthropic_llm import handler as anthropic_llm
from .gemini_llm import handler as gemini_llm
from .openai_llm import handler as openai_llm
from .togetherai_llm import handler as togetherai_llm


async def handler(
    # This is not actually used inside the function. It's a hack to have it as an argument
    # so we can use it when mapping data in the flow.
    prompt: Message = Field(default=None, description="Message to send to the LLM."), # noqa: ARG001
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    memory: list[Message] = Field(description="Chat history containing user and LLM messages."),
    model: str = Field(default="gpt-4o", description="Name of the LLM model to use."),
    tools: list[Tool | dict] = Field(default=None, description=" List of tools/functions the LLM can call."),
    tool_choice: dict[str, Any] | str = Field(
        default={"type": "auto"},
        description="How the model should use the provided tools"
    ),
    max_tokens: int = Field(
        default=1024,
        description="The maximum number of tokens in the response"
    ),
    temperature: float = Field(
        default=1,
        description=(
            "Sampling temperature (0-2 except for Anthropic which is 0-1). "
            "Higher values increase randomness, lower values increase determinism."
        )
    ),
    frequency_penalty: float = Field(
        default=0,
        description=(
            "Positive values penalize token frequency to reduce repetition. "
            "Ranges from -2.0 to 2.0."
        )
    ),
    presence_penalty: float = Field(
        default=0,
        description=(
            "Positive values penalize tokens based on presence to encourage new topics. "
            "Ranges from -2.0 to 2.0."
        )
    ),
    top_p: float = Field(
        default=1,
        description=(
            "Nucleus sampling parameter. Only tokens with cumulative probability "
            "mass up to top_p are considered."
        )
    ),
    top_k: int = Field(
        default=None,
        description="Only sample from the top K options for each subsequent token."
    ),
    logprobs: bool = Field(
        default=False,
        description="Whether to return logprobs with the returned text."
    ),
    top_logprobs: int = Field(
        default=None,
        description=(
            "Return log probabilities of the top N tokens (0-20). "
            "Requires logprobs=true."
        )
    ),
    seed: int = Field(
        default=None,
        description=(
            "Beta feature for deterministic sampling. Same seed and parameters "
            "should return same result."
        )
    ),
    stop: str | list[str] = Field(
        default=None,
        description="Where the model will stop generating."
    ),
    parallel_tool_calls: bool = Field(
        default=True,
        description="Whether to execute tool calls in parallel or sequentially."
    ),
    json_schema: dict = Field(
        default=None,
        description="JSON schema for structured output."
    ),
) -> Message: # type: ignore
    """Route requests to appropriate LLM providers based on model name prefix.

    This gateway function handles routing to different LLM providers (OpenAI, Anthropic,
    Gemini, TogetherAI) based on the model name prefix.

    Args:
        prompt: The input text to send to the LLM.
        memory: Chat history containing user and LLM messages.
        system_prompt: Instructions for the LLM to follow.
        model: Name of the LLM model to use.
        tools: List of available tool functions.
        tool_choice: Specification for tool selection.
        max_tokens: Maximum number of tokens in the response.
        temperature: Sampling temperature.
        frequency_penalty: Penalty for token frequency.
        presence_penalty: Penalty for token presence.
        top_p: Nucleus sampling parameter.
        top_k: Only sample from the top K options for each subsequent token.
        logprobs: Whether to return logprobs with the returned text.
        top_logprobs: Return log probabilities of the top N tokens.
        seed: Deterministic sampling parameter.
        stop: Up to 4 sequences where the model will stop generating.
        parallel_tool_calls: Whether to execute tool calls in parallel.
        json_schema: JSON schema for structured output.

    Yields:
        Any: Response chunks from the LLM provider
    """
    
    # Get default values for parameters
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    model = model.default if hasattr(model, 'default') else model
    tools = tools.default if hasattr(tools, 'default') else tools
    tool_choice = tool_choice.default if hasattr(tool_choice, 'default') else tool_choice
    max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    frequency_penalty = frequency_penalty.default if hasattr(frequency_penalty, 'default') else frequency_penalty
    top_logprobs = top_logprobs.default if hasattr(top_logprobs, 'default') else top_logprobs
    presence_penalty = presence_penalty.default if hasattr(presence_penalty, 'default') else presence_penalty
    seed = seed.default if hasattr(seed, 'default') else seed
    stop = stop.default if hasattr(stop, 'default') else stop
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    top_k = top_k.default if hasattr(top_k, 'default') else top_k
    parallel_tool_calls = parallel_tool_calls.default if hasattr(parallel_tool_calls, 'default') else parallel_tool_calls
    json_schema = json_schema.default if hasattr(json_schema, 'default') else json_schema

    # Route to appropriate provider based on model prefix
    if model.startswith("claude"):
        response = await anthropic_llm(
            prompt, system_prompt, memory, model, tools, tool_choice,
            max_tokens, stop, temperature, top_k, top_p, json_schema
        )
    elif model.startswith("gpt"):
        response = await openai_llm(
            prompt, system_prompt, memory, model, tools, tool_choice,
            max_tokens, frequency_penalty, logprobs, top_logprobs, presence_penalty,
            seed, stop, temperature, top_p, parallel_tool_calls, json_schema
        )
    elif model.startswith("gemini"):
        response = await gemini_llm(
            prompt, system_prompt, memory, model, tools, tool_choice,
            max_tokens, presence_penalty, seed, stop, temperature, top_p, json_schema
        )
    else:
        response = await togetherai_llm(
            prompt, system_prompt, memory, model, tools, tool_choice,
            max_tokens, frequency_penalty, top_logprobs, presence_penalty,
            seed, stop, temperature, top_p, parallel_tool_calls, json_schema
        )

    async for chunk in response:
        yield chunk



import inspect
import json

import pytest
from openai.types.chat import ChatCompletionChunk
from timbal.steps.llms.gateway import handler
from timbal.types.file import File
from timbal.types.models import create_model_from_argspec


@pytest.mark.asyncio
async def test_basic_response():
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is 2+2?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "meta-llama/Meta-Llama-3-8B-Instruct-Turbo"
    }
    model_args = model.model_validate(model_args)
        
    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
       if chunk.choices and chunk.choices[0].delta.content:
           response_chunks += chunk.choices[0].delta.content or ""

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "4" in response_chunks, "Response should contain the number 4"


@pytest.mark.asyncio
async def test_image_upload():
    url = "https://content.timbal.ai/assets/Three-Australian-Shepherd-puppies-sitting-in-a-field.jpg"
    file = File.validate(url)
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is in this image?"}, {"type": "file", "file": file}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "meta-llama/Llama-3.2-11B-Vision-Instruct-Turbo"
    }
    model_args = model.model_validate(model_args)

    response_chunks = ""
    async for chunk in handler(**dict(model_args)): 
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or ""

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "dog" in response_chunks.lower() or "pup" in response_chunks.lower(), \
        "Response should mention that there are dogs or puppies in the image"


@pytest.mark.asyncio
async def test_tools():
    tools = [{
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather for a location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The location to get weather for"
                    }
                },
                "required": ["location"]
            }
        }
    }]

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "What's the weather in London?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "tools": tools,
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = []
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, ChatCompletionChunk):
            # Handle tool calls
            if chunk.choices and chunk.choices[0].delta.tool_calls:
                tool_call = chunk.choices[0].delta.tool_calls[0]
                # If this is a new tool call, create a new entry
                if tool_call.id is not None:
                    current_tool_call = {
                        "type": "tool_use",
                        "id": tool_call.id,
                        "name": tool_call.function.name,
                        "input": ""
                    }
                    response_chunks.append(current_tool_call)
                # Accumulate arguments if present
                if tool_call.function.arguments:
                    current_tool_call["input"] += tool_call.function.arguments
            # Handle regular text content
            elif chunk.choices and chunk.choices[0].delta.content:
                if not response_chunks or "text" not in response_chunks[-1]:
                    response_chunks.append({"type": "text", "text": ""})
                response_chunks[-1]["text"] += chunk.choices[0].delta.content

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, list), "Response should be a list"
    weather_tool = next(
        (chunk for chunk in response_chunks if chunk.get("type") == "tool_use" and chunk.get("name") == "get_current_weather"), None
    )
    assert weather_tool is not None, "Weather tool call not found"
    assert isinstance(weather_tool["id"], str) and weather_tool["id"], "Tool call should have a non-empty ID"
    assert isinstance(weather_tool["input"], str), "Tool input should be a string"
    assert "London" in weather_tool["input"], "Tool input should contain 'London'"


@pytest.mark.asyncio
async def test_multiple_tools():
    tools= [{
        "type": "function",
        "function": {
            "name": "get_weather",
            "description": "Get the weather for a location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The location to get weather for"
                    }
                },
                "required": ["location"]
            }
        }
    },
    {"type": "function",
        "function": {
            "name": "get_time",
            "description": "Get the current time in a given time zone",
            "parameters": {
                "type": "object",
                "properties": {
                    "timezone": {
                        "type": "string",
                        "description": "The IANA time zone name, e.g. America/Los_Angeles"
                    }
                },
                "required": ["timezone"]
            }
        }}
    ]

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "What is the weather like right now in New York? Also what time is it there?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant. When a user asks multiple questions that require different tools, make sure to use all necessary tools to provide complete information.",
        "model": "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "tools": tools,
        "tool_choice": {"type":"auto"},
    }
    model_args = model.model_validate(model_args)
   
    response_chunks = []
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, ChatCompletionChunk):
            # Handle tool calls
            if chunk.choices and chunk.choices[0].delta.tool_calls:
                tool_call = chunk.choices[0].delta.tool_calls[0]
                # If this is a new tool call, create a new entry
                if tool_call.id is not None:
                    current_tool_call = {
                        "type": "tool_use",
                        "id": tool_call.id,
                        "name": tool_call.function.name,
                        "input": ""
                    }
                    response_chunks.append(current_tool_call)
                # Accumulate arguments if present
                if tool_call.function.arguments:
                    current_tool_call["input"] += tool_call.function.arguments
            # Handle regular text content
            elif chunk.choices and chunk.choices[0].delta.content:
                if not response_chunks or "text" not in response_chunks[-1]:
                    response_chunks.append({"type": "text", "text": ""})
                response_chunks[-1]["text"] += chunk.choices[0].delta.content

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, list), "Response should be a list"
    assert any(
        msg["type"] == "tool_use" and msg["name"] == "get_weather" 
        for msg in response_chunks
    ), "Weather tool call not found"
    assert any(
        msg["type"] == "tool_use" and msg["name"] == "get_time" 
        for msg in response_chunks
    ), "Time tool call not found"


@pytest.mark.asyncio
async def test_response_format():
    json_schema = {
        "title": {"type": "string"},
        "author": {"type": "string"},
        "genre": {"type": "string"},
        "publication_year": {"type": "integer"},
        "summary": {"type": "string"}
    }

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "Summarize The Eras Tour by Taylor Swift with fields for title, author, genre, publication year, and a short summary."}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "system_prompt": "Respond in JSON format",
        "memory": [memory],
        "model": "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "json_schema": json_schema
    }
    model_args = model.model_validate(model_args)

    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or ""  
    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert json.loads(response_chunks), "Response should match the JSON schema"


import logging
import os

import structlog


def setup_logging() -> None:
    """Configures the logging system using `structlog`.

    The function uses the TIMBAL_LOG_LEVEL and TIMBAL_LOG_FORMAT environment variables 
    to determine the logging level and format. It supports both human-readable and 
    JSON-based structured logging.
    """
    log_level = os.getenv("TIMBAL_LOG_LEVEL", "INFO")

    # Switch to human readable log output if TIMBAL_LOG_FORMAT is set to "dev"
    dev_logs = os.getenv("TIMBAL_LOG_FORMAT", "") == "dev"

    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.TimeStamper(fmt="iso"),
    ]

    if dev_logs:
        processors.append(structlog.dev.set_exc_info)
    else:
        processors.append(structlog.processors.format_exc_info)

    structlog.configure(
        processors=processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    if dev_logs:
        log_renderer = structlog.dev.ConsoleRenderer(event_key="message")
    else:
        log_renderer = structlog.processors.JSONRenderer()

    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=processors,
        processors=[
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(log_level)



import pytest
from timbal import Flow
from timbal.state.context import RunContext
from timbal.state.savers import InMemorySaver


@pytest.mark.asyncio
async def test_invalid_run_id():
    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    context = RunContext(id="1")

    prompt = "Hello my name is David"
    await flow.complete(context=context, prompt=prompt)

    with pytest.raises(ValueError):
        prompt = "What is my name?"
        await flow.complete(context=context, prompt=prompt)


@pytest.mark.asyncio
async def test_no_parent_id():
    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    prompt = "Hello my name is David"
    await flow.complete(prompt=prompt)

    prompt = "What is my name?"
    async for event in flow.run(prompt=prompt):
        if event.type == "FLOW_OUTPUT":
            assert "david" not in event.output["message"].content[0].text.lower()


@pytest.mark.asyncio
async def test_parent_id():
    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    prompt = "Hello my name is David"
    flow_output_event = await flow.complete(prompt=prompt)

    prompt = "What is my name?"
    async for event in flow.run(
        context=RunContext(parent_id=flow_output_event.run_id),
        prompt=prompt,
    ):
        if event.type == "FLOW_OUTPUT":
            assert "david" in event.output["message"].content[0].text.lower()


@pytest.mark.asyncio
async def test_different_parent_id():

    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    prompt = "Hello my name is David"
    await flow.complete(prompt=prompt)

    prompt = "What is my name?"
    async for event in flow.run(
        context=RunContext(parent_id="wrong_parent_id"),
        prompt=prompt,
    ):
        if event.type == "FLOW_OUTPUT":
            assert "david" not in event.output["message"].content[0].text.lower()


@pytest.mark.asyncio
async def test_rewind():

    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    prompt = "Hello"
    flow_output_event_1 = await flow.complete(prompt=prompt)

    prompt = "My name is David"
    await flow.complete(
        context=RunContext(parent_id=flow_output_event_1.run_id),
        prompt=prompt,
    )

    prompt = "What is my name?"
    async for event in flow.run(
        context=RunContext(parent_id=flow_output_event_1.run_id),
        prompt=prompt,
    ):
        if event.type == "FLOW_OUTPUT":
            assert "david" not in event.output["message"].content[0].text.lower()


@pytest.mark.asyncio
async def test_rewind_2():
    flow = (
        Flow(id="flow")
        .add_llm(id="llm", memory_id="llm")
        .set_data_map("llm.prompt", "prompt")
        .set_output("llm.return", "message")
        .compile(state_saver=InMemorySaver())
    )

    prompt = "Hello"
    flow_output_event_1 = await flow.complete(prompt=prompt)

    prompt = "My name is David"
    flow_output_event_2 = await flow.complete(
        context=RunContext(parent_id=flow_output_event_1.run_id),
        prompt=prompt,
    )

    prompt = "What is my name?"
    flow_output_event_3 = await flow.complete(
        context=RunContext(parent_id=flow_output_event_2.run_id), 
        prompt=prompt,
    )
    assert "david" in flow_output_event_3.output["message"].content[0].text.lower()

    prompt = "My name is John"
    flow_output_event_4 = await flow.complete(
        context=RunContext(parent_id=flow_output_event_1.run_id),
        prompt=prompt,
    )

    prompt = "What is my name?"
    flow_output_event_5 = await flow.complete(
        context=RunContext(parent_id=flow_output_event_4.run_id),
        prompt=prompt,
    )
    assert "john" in flow_output_event_5.output["message"].content[0].text.lower()



import pytest
from timbal import Flow
from timbal.types import Field


def get_weather(
    location: str = Field(description="The location to get the weather for"), # noqa: ARG001
    unit: str = Field(choices=["celsius", "fahrenheit"], description="The unit for the temperature") # noqa: ARG001
) -> str:
    return "sunny"


def get_traffic(
    location: str = Field(description="The location to get the weather for"), # noqa: ARG001
) -> str:
    return "pretty heavy"


@pytest.mark.asyncio
async def test_tool_use():
    flow = (
        Flow(id="test_parallel_tools")
        .add_llm("ask_llm")
        .set_data_map("ask_llm.prompt", "prompt")
        .add_step("get_weather", get_weather)
        .add_link("ask_llm", "get_weather", is_tool=True)
        .add_step("get_traffic", get_traffic)
        .add_link("ask_llm", "get_traffic", is_tool=True)
    )

    prompt = "What's the weather and traffic like in New York?"
    executed_steps = set()
    async for event in flow.run(prompt=prompt):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "get_weather" in executed_steps
    assert "get_traffic" in executed_steps
    
    # Use only the tool that is relevant to the prompt
    prompt = "What's the weather like in New York?"
    executed_steps.clear()
    async for event in flow.run(prompt=prompt):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "get_weather" in executed_steps
    assert "get_traffic" not in executed_steps

    # No tools should be used
    prompt = "What's 2 + 2?"
    executed_steps.clear()
    async for event in flow.run(prompt=prompt):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "get_weather" not in executed_steps
    assert "get_traffic" not in executed_steps


@pytest.mark.asyncio
async def test_tool_result():
    flow = (
        Flow(id="test_tool_result")
        .add_llm("ask_llm", memory_id="ask_llm")
        .set_data_map("ask_llm.prompt", "prompt")
        .add_step("get_weather", get_weather)
        .add_link("ask_llm", "get_weather", is_tool=True)
        .add_step("get_traffic", get_traffic)
        .add_link("ask_llm", "get_traffic", is_tool=True)
        .add_llm("ask_llm_2", memory_id="ask_llm")
        .add_link("get_weather", "ask_llm_2", is_tool_result=True)
        .add_link("get_traffic", "ask_llm_2", is_tool_result=True)
    )

    prompt = "What's the weather and traffic like in New York?"
    executed_steps = set()
    async for event in flow.run(prompt=prompt):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "get_weather" in executed_steps
    assert "get_traffic" in executed_steps
    assert "ask_llm_2" in executed_steps



# ruff: noqa: B008
import inspect
from typing import Annotated, Any, Literal, Optional, TypeVar

import pytest
from pydantic import ValidationError
from timbal import Flow
from timbal.types import Field, File
from timbal.types.models import create_model_from_annotation, create_model_from_argspec


def test_no_annotations_no_defaults():
    def handler(a, b, c): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    handler_args = model_cls.model_validate({"a": 1, "b": 2, "c": 3})
    assert dict(handler_args) == {"a": 1, "b": 2, "c": 3}

    with pytest.raises(ValueError):
        model_cls.model_validate({"a": 1, "b": 2})


def test_no_defaults():
    def handler(a, b: str, c: float): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    handler_args = model_cls.model_validate({"a": 1, "b": "2", "c": 3.0})
    assert dict(handler_args) == {"a": 1, "b": "2", "c": 3.0}

    with pytest.raises(ValidationError):
        model_cls.model_validate({"a": 1, "b": "2"})

    with pytest.raises(ValidationError):
        model_cls.model_validate({"a": 1, "b": "2", "c": "this type is wrong"})


def test_choices():
    def handler(
        a: Literal["a", "b", "c"],
        b: str = Field(choices=["a", "b", "c"])
    ): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    handler_args = model_cls.model_validate({"a": "a", "b": "b"})
    assert dict(handler_args) == {"a": "a", "b": "b"}

    with pytest.raises(ValidationError):
        model_cls.model_validate({"a": "a", "b": "d"})

    with pytest.raises(ValidationError):
        model_cls.model_validate({"a": "f", "b": "b"})
    

def test_regex():
    def handler(a: str = Field(regex=r"^[a-z]+$")): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    model_args = {"a": "abc"}
    model = model_cls(**model_args)
    assert model.model_dump() == {"a": "abc"}

    model_args = {"a": "123"}
    with pytest.raises(ValidationError):
        model_cls(**model_args)


def test_optionals():
    def handler(
        a: Optional[str], # noqa: UP007
        b: float | None, 
        c: int | None = None,
        d: list[int] | None = None,
    ): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    model_args = {"a": "a", "b": 1.2, "c": 3, "d": [1, 2, 3]}
    model = model_cls(**model_args)
    assert model.model_dump() == {"a": "a", "b": 1.2, "c": 3, "d": [1, 2, 3]}

    # With missing fields
    model_cls.model_validate({"a": "a", "b": 1.2})

    # Wrong list item
    with pytest.raises(ValidationError):
        model_cls.model_validate({"a": "a", "b": 1.2, "d": [1, 2, "abc"]})

    # Passing None
    model_cls.model_validate({"a": None, "b": None, "c": None, "d": None})


def test_generics():
    T1 = TypeVar("T1")
    T2 = TypeVar("T2")
    def handler(
        a: list[int], 
        b: list[T1], 
        c: list[T2] = Field(), 
        d: list[list[T2]] = Field()
    ): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    model_args = {"a": [1, 2, 3], "b": [6.7, 8.99], "c": [], "d": [["d"]]}
    model = model_cls(**model_args)
    assert model.model_dump() == {"a": [1, 2, 3], "b": [6.7, 8.99], "c": [], "d": [["d"]]}

    # list[int] must be a list of ints
    model_args = {"a": ["a"], "b": [], "c": [], "d": [["d"]]}
    with pytest.raises(ValueError):
        model_cls(**model_args)

    # list[T] must be a list of the same type
    model_args = {"a": [1], "b": ["a", "b", 1], "c": [], "d": [["d"]]}
    with pytest.raises(ValueError):
        model_cls(**model_args)

    # list[T] must be a list of the same type between fields
    model_args = {"a": [1], "b": [1.2, 3.4], "c": [1], "d": [["d"]]}
    with pytest.raises(ValueError):
        model_cls(**model_args)


def test_file():
    def handler(a: File): pass

    argspec = inspect.getfullargspec(handler)
    model_cls = create_model_from_argspec("TestModel", argspec)

    file_url = "https://content.eticko.com/assets/timbal.png"

    model_args = {"a": file_url}
    model = model_cls(**model_args)

    file = File.validate(file_url)
    file_serialized = File.serialize(file)
    assert model.model_dump() == {"a": file_serialized}


# def test_message():
#     def handler(a: Message): pass

#     argspec = inspect.getfullargspec(handler)
#     model_cls = create_model_from_argspec("TestModel", argspec)

#     model_cls.model_validate({"a": {"role": "user", "content": "hello"}})
#     model_cls.model_validate({"a": Message.validate({"role": "user", "content": "hello"})})

    # from timbal import Flow

    # flow = (
    #     Flow(id="test_flow_message")
    #     .add_llm("llm")
    #     .set_data_map("llm.prompt", "prompt")
    # )

    # prompt = "hello"
    # prompt = File.validate("https://content.eticko.com/assets/timbal.png")
    # prompt = {"type": "text", "text": "hello"}
    # prompt = [File.validate("https://content.eticko.com/assets/timbal.png"), "describe the shape of the logo"]
    # prompt = [
    #     {"type": "file", "file": File.validate("https://content.eticko.com/assets/timbal.png")}, 
    #     "describe the shape of the logo",
    #     "what do you think the company behind the logo is about?",
    #     "invent the background story of the company"
    # ]
    # prompt = {
    #     "role": "user",
    #     "content": "how are you in this fine day?"
    # }
    # prompt = Message.validate({"role": "user", "content": "how are you in this fine day?"})
    # async for event_code, event_data in flow.run(prompt=prompt):
    #     print(event_code, event_data)


def test_return_model():
    def handler(a) -> int:
        return a

    return_annotation = inspect.getfullargspec(handler).annotations.get("return", ...)
    return_model = create_model_from_annotation("TempModel", return_annotation)

    return_model.model_validate({"return": handler(1)})

    with pytest.raises(ValidationError):
        return_model.model_validate({"return": handler("a")})


def test_return_model_any():
    def handler(a) -> Any:
        return a

    return_annotation = inspect.getfullargspec(handler).annotations.get("return", ...)
    return_model = create_model_from_annotation("TempModel", return_annotation)

    return_model.model_validate({"return": handler(23)})


def test_return_model_no_annotation():
    def handler(a):
        return a

    return_annotation = inspect.getfullargspec(handler).annotations.get("return", ...)
    return_model = create_model_from_annotation("TempModel", return_annotation)

    return_model.model_validate({"return": handler({})})


def test_return_model_none():
    def handler(a) -> None:
        pass

    return_annotation = inspect.getfullargspec(handler).annotations.get("return", ...)
    return_model = create_model_from_annotation("TempModel", return_annotation)

    return_model.model_validate({"return": handler(None)})


def test_return_model_annotated():
    def handler(a) -> Annotated[int, "this is a test"]:
        return a

    return_annotation = inspect.getfullargspec(handler).annotations.get("return", ...)
    return_model = create_model_from_annotation("TempModel", return_annotation)

    return_model.model_validate({"return": handler(1)})


def test_flow_return_model():
    def handler_1(a, b) -> int: pass
    def handler_2(a) -> str: pass
    def handler_3(a) -> list[str]: pass
    def handler_4(a) -> dict[str, int]: pass
    def handler_5(a) -> Any: pass

    flow = (
        Flow(id="test_flow_return_model")
        .add_step("1", handler_1)
        .add_step("2", handler_2)
        .add_step("3", handler_3)
        .add_step("4", handler_4)
        .add_step("5", handler_5)
        .set_data_value("something", ["hello", "world"])
        .set_output("1.return", "out_1")
        .set_output("2.return", "out_2")
        .set_output("3.return", "out_3")
        .set_output("4.return.abc", "out_4")
        .set_output("5.return", "out_5")
        .set_output("something.1", "out_6")
    )

    flow_return_model = flow.return_model()
    flow.return_model_schema() # Simply ensure this doesn't raise an error

    flow_return_model.model_validate({
        "out_1": 1,
        "out_2": "world",
        "out_3": ["hello", "world"],
        "out_4": {"abc": 123},
        "out_5": 34.12,
        "out_6": "world",
    })
    


from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..state import RunContext


class BaseStep(BaseModel, ABC):
    """Abstract base class for defining processing steps in a workflow.

    BaseStep combines Pydantic's data validation with abstract methods to create a
    standardized interface for workflow steps. Each step must define its parameter
    and return value schemas, as well as the actual processing logic.
    """
    # Allow storing extra fields in the model.
    model_config = ConfigDict(extra="allow")

    id: str 
    """Unique identifier for the step instance."""
    path: str 
    """Any step will be a part of a flow. With potentially multiple nested sub-flows.
    We will use the path to uniquely identify the step's position in the overall flow.
    """
    metadata: dict[str, Any] = {}
    """Optional metadata associated with the step."""


    @abstractmethod
    def params_model(self) -> BaseModel:
        """Returns the Pydantic model defining the expected parameters for this step."""
        pass


    @abstractmethod
    def params_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the step's parameter model."""
        pass
    

    @abstractmethod
    def return_model(self) -> Any:
        """Returns the expected return type for this step."""
        pass


    @abstractmethod
    def return_model_schema(self) -> dict[str, Any]:
        """Returns the JSON schema for the step's return value model."""
        pass


    # TODO Better method definition. Then we can use "See base class" in the child classes.
    @abstractmethod
    def run(
        self, 
        context: RunContext | None = None, # noqa: ARG002
        **kwargs: Any,
    ) -> Any:
        """Executes the step's processing logic."""
        pass



import json
from pathlib import Path

from pydantic import TypeAdapter

from ...types.models import dump
from ..context import RunContext
from ..data import Data
from ..snapshot import Snapshot
from .base import BaseSaver


class JSONLSaver(BaseSaver):
    """A JSONL state saver.

    This state saver stores snapshots in a JSONL file by serializing every snapshot into every line.

    Note:
        Only use `JSONLSaver` for debugging or testing purposes.
        This saver was implemented to test serialization and deserialization of snapshots.
        For production use cases, use a persistent state saver like `PostgresSaver`.
    """

    def __init__(self, path: Path | str) -> None:
        """Initialize a JSONLSaver instance.

        Args: 
            path: Path to the JSONl file that will store the snapshots. 
        """
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"'path' must be a string or a Path, got {type(path)}.")
        self.path = path.expanduser().resolve()
        # Ensure the directory exists
        self.path.parent.mkdir(parents=True, exist_ok=True)
        # Create the file if it doesn't exist
        if not self.path.exists():
            self.path.touch()


    @staticmethod 
    def _load_snapshot_from_line(line: str) -> Snapshot:
        snapshot = json.loads(line.strip())
        snapshot["data"] = {
            k: TypeAdapter(Data).validate_python(v)
            for k, v in snapshot["data"].items()
        }
        return Snapshot(**snapshot)

    
    def get_last(
        self, 
        path: str,
        context: RunContext,
    ) -> Snapshot | None:
        """See base class.

        Warning:
            This method loads the entire file into memory. For production use cases with large files,
            consider implementing a streaming approach that reads the file line by line.
        """
        if context.parent_id is None:
            return None 

        with open(self.path) as f:
            for line in reversed(list(f)):
                snapshot = self._load_snapshot_from_line(line)

                if snapshot.path == path and snapshot.id == context.parent_id:
                    return snapshot

        return None
    

    def put(
        self, 
        snapshot: Snapshot,
        context: RunContext, # noqa: ARG002
    ) -> None:
        """See base class."""
        # Since we're appending lines to a file and there's no intrinsic way of ensuring
        # unicity of ids, we need to check if the snapshot already exists.
        with open(self.path) as f:
            for line in reversed(list(f)):
                snapshot_i = self._load_snapshot_from_line(line)
                if snapshot_i.id == snapshot.id and snapshot_i.path == snapshot.path:
                    raise ValueError(f"Snapshot with id {snapshot.id} and path {snapshot.path} already exists.")

        with open(self.path, "a") as f:
            snapshot_dump = dump(snapshot)
            f.write(json.dumps(snapshot_dump) + "\n")
    


from typing import Any

import pytest
from timbal import Flow
from timbal.types import Field


def identity_handler(x: Any = Field(default=None)) -> Any:
    return x


@pytest.mark.asyncio
async def test_remove_step():
    flow = (
        Flow(id="test_remove_step")
        .add_step("step_1", identity_handler)
        .add_step("step_2", identity_handler)
        .add_step("step_3", identity_handler)
        .set_data_map("step_1.x", "x1")
        .set_data_map("step_2.x", "x2")
        .set_output("step_1.return", "x1")
        .set_output("step_2.return", "x2")
        .set_data_map("step_3.x", "step_1.return")
    )
    
    flow.remove_step("step_1")

    assert "step_1" not in flow.steps

    # Ensure we're cleaning up links correctly
    assert "step_1-step_2" not in flow.links

    # Ensure we're cleaning up outputs correctly
    assert "x1" not in flow.outputs
    assert "x2" in flow.outputs

    # Ensure we're cleaning up data correctly
    assert "step_1.x" not in flow.data
    assert "step_2.x" in flow.data
    assert "step_3.x" not in flow.data



import pathlib
from io import BytesIO

import pytest
from timbal.types import File


def test_file_validate_none() -> None:
    with pytest.raises(ValueError):
        File.validate(None)


def test_file_validate_any() -> None:
    with pytest.raises(ValueError):
        File.validate(object())


def test_file_serialize_none() -> None:
    assert File.serialize(None) is None


def test_file_serialize_any() -> None:
    with pytest.raises(ValueError):
        File.serialize(object())


def test_file_validate_bytes() -> None:
    test_content = b"Hello, World!"
    file = File.validate(test_content)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "bytes"
    assert file.__source_extension__ is None
    assert file.read() == test_content


def test_file_validate_bytearray() -> None:
    test_content = bytearray(b"Hello, World!")
    file = File.validate(test_content)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "bytes"
    assert file.__source_extension__ is None
    assert file.read() == bytes(test_content)


def test_file_validate_io_base() -> None:
    test_content = b"Hello, World!"
    test_io = BytesIO(test_content)
    file = File.validate(test_io)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "bytes"
    assert file.__source_extension__ is None
    assert file.read() == test_content


def test_file_serialize_byteslike() -> None:
    test_content = b"Hello, World!"
    test_io = BytesIO(test_content)

    file = File.validate(test_io)

    assert File.serialize(file) == "data:application/octet-stream;base64,SGVsbG8sIFdvcmxkIQ=="


def test_file_validate_local_path_from_str(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "test.txt"
    test_content = b"Hello, World!"
    test_file.write_bytes(test_content)

    file = File.validate(str(test_file))

    assert isinstance(file, File)
    assert file.__source_scheme__ == "local_path"
    assert file.__source_extension__ == test_file.suffix
    assert file.read() == test_content


def test_file_validate_local_path_from_path(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "test.txt"
    test_content = b"Hello, World!"
    test_file.write_bytes(test_content)

    file = File.validate(test_file)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "local_path"
    assert file.__source_extension__ == test_file.suffix
    assert file.read() == test_content


def test_file_serialize_local_path(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "test.txt"
    test_content = b"Hello, World!"
    test_file.write_bytes(test_content)

    file = File.validate(str(test_file))

    assert File.serialize(file) == "data:text/plain;base64,SGVsbG8sIFdvcmxkIQ=="


def test_file_validate_data_url() -> None:
    test_content = b"Hello, World!"
    data_url = "data:text/plain;base64,SGVsbG8sIFdvcmxkIQ=="

    file = File.validate(data_url)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "data_url"
    assert file.__source_extension__ == ".txt"
    assert file.read() == test_content


def test_file_validate_data_url_with_custom_mime_type() -> None:
    test_content = b'{"key": "value"}'
    data_url = "data:timbal/jsonl;base64,eyJrZXkiOiAidmFsdWUifQ=="

    file = File.validate(data_url)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "data_url"
    assert file.__source_extension__ == ".jsonl"
    assert file.read() == test_content


def test_file_serialize_data_url() -> None:
    data_url = "data:timbal/jsonl;base64,eyJrZXkiOiAidmFsdWUifQ=="

    file = File.validate(data_url)

    assert File.serialize(file) == data_url


def test_file_validate_url() -> None:
    test_content = b"Hello, World!\n"
    url = "https://content.timbal.ai/assets/hello_world.txt"

    file = File.validate(url)

    assert isinstance(file, File)
    assert file.__source_scheme__ == "url"
    assert file.__source_extension__ == ".txt"
    assert file.read() == test_content


def test_file_serialize_url() -> None:
    url = "https://content.timbal.ai/assets/hello_world.txt"

    file = File.validate(url)

    assert File.serialize(file) == "data:text/plain;base64,SGVsbG8sIFdvcmxkIQo="



import argparse
import json
import os
import sys
from pathlib import Path

import structlog
from dotenv import load_dotenv

from .. import Flow, __version__
from ..logs import setup_logging
from .utils import ModuleSpec, load_module

logger = structlog.get_logger("timbal.server.probe")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Timbal probe script.")
    parser.add_argument(
        "-v", 
        "--version", 
        action="store_true", 
        help="Show version and exit."
    )
    parser.add_argument(
        "--module_spec",
        dest="module_spec",
        type=str,
        help="Path to a python module and optional object (format: path/to/file.py::object_name)",
    )
    args = parser.parse_args()

    if args.version:
        print(f"timbal.servers.http {__version__}") # noqa: T201
        sys.exit(0)

    # We can overwrite the env TIMBAL_FLOW variable with the --module_spec flag.
    module_spec = args.module_spec
    if not module_spec:
        module_spec = os.getenv("TIMBAL_FLOW")

    if not module_spec:
        print("No module spec provided. Set TIMBAL_FLOW env variable or use --module_spec to specify a module to load.") # noqa: T201
        sys.exit(1)

    module_parts = module_spec.split(":")
    if len(module_parts) > 2:
        print("Invalid module spec format. Use 'path/to/file.py:object_name' or 'path/to/file.py'") # noqa: T201
        sys.exit(1)
    elif len(module_parts) == 2:
        module_path, module_name = module_parts
        module_spec = ModuleSpec(
            path=Path(module_path).expanduser().resolve(), 
            object_name=module_name,
        )
    else:
        module_spec = ModuleSpec(
            path=Path(module_parts[0]).expanduser().resolve(), 
            object_name=None,
        )

    load_dotenv()
    setup_logging()

    flow = load_module(module_spec)

    if not isinstance(flow, Flow):
        raise ValueError("The loaded module is not a valid Flow instance.")

    params_model_schema = flow.params_model_schema()
    return_model_schema = flow.return_model_schema()

    output = {
        "params_model_schema": params_model_schema,
        "return_model_schema": return_model_schema
    }
    print(json.dumps(output)) # noqa: T201



from typing import Any

import pytest
from timbal import Flow
from timbal.types import Field


def identity_handler(x: Any = Field(default=None)) -> Any:
    return x


@pytest.mark.asyncio
async def test_flow_conditions():
    flow = (
        Flow(id="test_flow_conditions")
        .add_step("step_1", identity_handler)
        .add_step("step_2", identity_handler)
        .add_link("step_1", "step_2", condition="{{step_1.return}} * {{x}} > 5")
        .set_data_map("step_1.x", "x")
    )

    executed_steps = set()
    async for event in flow.run(x=1):
        if event.type == "STEP_START": 
            executed_steps.add(event.step_id)
    assert "step_2" not in executed_steps

    executed_steps.clear()
    async for event in flow.run(x=3):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "step_2" in executed_steps


@pytest.mark.asyncio
async def test_flow_positive_negative_conditions():
    flow = (
        Flow(id="test_flow_positive_negative_conditions")
        .add_step("step_1", identity_handler)
        .add_step("step_2", identity_handler)
        .add_step("step_3", identity_handler)
        .add_link("step_1", "step_2", condition="{{step_1.return}} > 0")
        .add_link("step_1", "step_3", condition="{{step_1.return}} <= 0")
        .set_data_map("step_1.x", "x")
    )

    # Test positive path
    steps_executed = set()
    async for event in flow.run(x=1):
        if event.type == "STEP_START":
            steps_executed.add(event.step_id)
    assert "step_2" in steps_executed
    assert "step_3" not in steps_executed

    # Test negative path
    steps_executed.clear()
    async for event in flow.run(x=-1):
        if event.type == "STEP_START":
            steps_executed.add(event.step_id)
    assert "step_2" not in steps_executed
    assert "step_3" in steps_executed


@pytest.mark.asyncio
async def test_flow_string_conditions():
    flow = (
        Flow(id="test_flow_string_conditions")
        .add_step("input", identity_handler)
        .add_step("path_a", identity_handler)
        .add_step("path_b", identity_handler)
        .add_link("input", "path_a", condition="{{input.return}} == 'test'")
        .add_link("input", "path_b", condition="{{input.return}} != 'test'")
        .set_data_map("input.x", "input_str")
    )

    executed_steps = set()
    async for event in flow.run(input_str="test"):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "path_a" in executed_steps
    assert "path_b" not in executed_steps


@pytest.mark.asyncio
async def test_flow_list_conditions():
    flow = (
        Flow(id="test_flow_list_conditions")
        .add_step("list_step", identity_handler)
        .add_step("int_step", identity_handler)
        .add_link("list_step", "int_step", condition="{{list_step.return}} == ['TEST']")
        .set_data_map("list_step.x", "input_list")
    )

    executed_steps = set()
    async for event in flow.run(input_list=["TEST"]):
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "int_step" in executed_steps


@pytest.mark.asyncio
async def test_many_to_one():
    flow = (
        Flow(id="test_many_to_one")
        .add_step("step_1", identity_handler)
        .add_step("step_2", identity_handler)
        .add_step("step_3", identity_handler)
        .add_link("step_1", "step_3", condition="1 > 0")
        .add_link("step_2", "step_3", condition="1 < 0")
    )

    executed_steps = set()
    async for event in flow.run():
        if event.type == "STEP_START":
            executed_steps.add(event.step_id)
    assert "step_3" in executed_steps



from . import errors, state, steps, types
from .graph import Agent, Flow

try:
    from ._version import __version__  # type: ignore
except ImportError:
    __version__ = "0.0.0.dev0"


__all__ = [
    "__version__",
    "Agent",
    "Flow",
    "errors",
    "state",
    "steps",
    "types",
]



"""
Utilities for working with Directed Acyclic Graphs (DAGs).

This module provides functions to analyze and traverse DAGs represented as adjacency lists.
Each graph is represented as a dictionary mapping node IDs to sets of their direct successors.

Example:
    dag = {
        'A': {'B', 'C'},
        'B': {'D'},
        'C': {'D'},
        'D': set()
    }
"""

# Type alias for a Directed Acyclic Graph (stored in adjacency list format).
Dag = dict[str, set[str]]


def reverse_dag(dag: Dag) -> Dag:
    """Function to reverse a Directed Acyclic Graph (DAG)."""
    return {
        node_id: {prev_node_id for prev_node_id in dag if node_id in dag[prev_node_id]} 
        for node_id in dag
    }


def is_dag(graph: Dag) -> bool:
    """Function to check if a graph is a Directed Acyclic Graph (DAG)."""
    # States: 0 = unvisited, 1 = visiting, 2 = visited
    state = {node_id: 0 for node_id in graph}
    def dfs(node_id):
        if state[node_id] == 1:
            return False
        if state[node_id] == 2:
            return True
        state[node_id] = 1
        for next_node_id in graph[node_id]:
            if not dfs(next_node_id):
                return False
        state[node_id] = 2
        return True
    for node_id in graph:
        if state[node_id] == 0:
            if not dfs(node_id):
                return False
    return True


def get_ancestors(node_id: str, rev_dag: Dag) -> set[str]:
    """Function to get the ancestors of a node in a (reversed) Directed Acyclic Graph (DAG)."""
    def dfs(node_id, visited):
        if node_id in visited:
            return
        visited.add(node_id)
        for prev_node_id in rev_dag[node_id]:
            dfs(prev_node_id, visited)
        return visited
    ancestors = dfs(node_id, set()) - {node_id}
    return ancestors


def get_successors(node_id: str, dag: Dag) -> set[str]:
    """Function to get the successors of a node in a Directed Acyclic Graph (DAG)."""
    def dfs(node_id, visited):
        if node_id in visited:
            return
        visited.add(node_id)
        for prev_node_id in dag[node_id]:
            dfs(prev_node_id, visited)
        return visited
    successors = dfs(node_id, set()) - {node_id}
    return successors


def get_sources(dag: Dag) -> set[str]:
    """Function to retrieve the nodes with no incoming edges of a Directed Acyclic Graph (DAG)."""
    all_nodes = set(dag)
    nodes_with_incoming_edges = set()
    for next_nodes_ids in dag.values():
        nodes_with_incoming_edges.update(next_nodes_ids)
    initial_steps_ids = all_nodes - nodes_with_incoming_edges
    return initial_steps_ids



from typing import Any

from anthropic.types import (
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
)
from openai.types.chat import ChatCompletionChunk
from pydantic import BaseModel, ConfigDict

# Create a type alias for Anthropic events
AnthropicEvent = (
    RawContentBlockStartEvent |
    RawContentBlockDeltaEvent |
    RawContentBlockStopEvent |
    RawMessageStartEvent |
    RawMessageDeltaEvent |
    RawMessageStopEvent
)


# Create a type alias for OpenAI events (in the future we may have more than one)
OpenAIEvent = ChatCompletionChunk


class TimbalEvent(BaseModel):
    """Base class for all timbal events yielded during flow execution."""
    # Allow storing extra fields in the model.
    model_config = ConfigDict(extra="allow")

    type: str
    """The type of the event. This will be very useful for serializing and deserializing events."""
    run_id: str | None = None
    """The id of the run this event was emitted from."""
    parent_step_id: str | None = None
    """The id of the parent step that yielded this event."""
    step_id: str | None = None
    """The id of the step that yielded this event."""


class StepStartEvent(TimbalEvent):
    """Event emitted when a step starts execution."""
    type: str = "STEP_START"


class StepChunkEvent(TimbalEvent):
    """Event emitted for streaming output chunks from a step."""
    type: str = "STEP_CHUNK"

    step_chunk: Any
    """The chunk of output from the step."""


class StepOutputEvent(TimbalEvent):
    """Event emitted when a step completes with its full output."""
    type: str = "STEP_OUTPUT"

    input: Any
    """The input arguments passed to the step."""
    output: Any | None = None
    """The result of the step (if any)."""
    error: Any | None = None
    """The error that occurred during the step (if any)."""
    t0: int 
    """The start time of the step in milliseconds."""
    t1: int 
    """The end time of the step in milliseconds."""
    usage: Any | None = None
    """The usage of the step."""


class FlowOutputEvent(TimbalEvent):
    """Event emitted when a flow completes with its full output."""
    type: str = "FLOW_OUTPUT"

    output: Any | None = None
    """The output of the flow."""



import base64
import inspect
import json
import uuid

import openai
import pytest
import requests
from openai.types.chat import ChatCompletionChunk
from timbal.steps.llms.gateway import handler
from timbal.types.file import File
from timbal.types.message import Message
from timbal.types.models import create_model_from_argspec


@pytest.mark.asyncio
async def test_basic_response():
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is 2+2?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "gemini-2.0-flash-lite-preview-02-05"
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = ""
    async for chunk in handler(**dict(model_args)): 
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or "" 

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "4" in response_chunks, "Response should contain the number 4"


@pytest.mark.asyncio
async def test_image_upload():
    url = "https://content.timbal.ai/assets/Three-Australian-Shepherd-puppies-sitting-in-a-field.jpg"
    response = requests.get(url)
    image_data = response.content
    base64_image = base64.b64encode(image_data).decode('utf-8')
    file = File.validate(f"data:image/jpeg;base64,{base64_image}")
    
    memory = {
        "role": "user", 
        "content": [{"type": "text", "text": "What is in this image?"}, {"type": "file", "file": file}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "gemini-2.0-flash-lite-preview-02-05"
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = ""
    async for chunk in handler(**dict(model_args)): 
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or ""   

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "dog" in response_chunks.lower() or "pup" in response_chunks.lower(), \
        "Response should mention that there are dogs or puppies in the image"


@pytest.mark.asyncio
async def test_audio_upload():
    url = "https://cdn.openai.com/API/docs/audio/alloy.wav"
    file = File.validate(url)
    
    memory = {
        "role": "user", 
        "content": [
            {"type": "text", "text": "What is in this recording?"}, 
            {"type": "file", "file": file}
        ]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)

    model_args = {
        "memory": [memory],
        "model": "gemini-2.0-flash-lite-preview-02-05"
    }
    model_args = model.model_validate(model_args)

    response_chunks = ""
    async for chunk in handler(**dict(model_args)): 
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or ""

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert "sun" in response_chunks.lower(), "Response should mention the sun"


@pytest.mark.asyncio
async def test_tools():
    tools = [{
        "type": "function",
        "function": {
            "name": "get_weather",
            "description": "Get the weather for a location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The location to get weather for"
                    }
                },
                "required": ["location"]
            }
        }
    }]

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "What's the weather in London?"}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "gemini-2.0-flash-lite-preview-02-05",
        "tools": tools,
        "tool_choice": {"type": "auto"}
    }
    model_args = model.model_validate(model_args)
    
    response_chunks = []
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, ChatCompletionChunk):
            # Handle tool calls
            if chunk.choices and chunk.choices[0].delta.tool_calls:
                tool_call = chunk.choices[0].delta.tool_calls[0]
                # If this is a new tool call, create a new entry
                if tool_call.id is not None:
                    generated_id = str(uuid.uuid4())
                    current_tool_call = {
                        "type": "tool_use",
                        "id": generated_id,
                        "name": tool_call.function.name,
                        "input": ""
                    }
                    response_chunks.append(current_tool_call)
                # Accumulate arguments if present
                if tool_call.function.arguments:
                    current_tool_call["input"] += tool_call.function.arguments
            # Handle regular text content
            elif chunk.choices and chunk.choices[0].delta.content:
                if not response_chunks or "text" not in response_chunks[-1]:
                    response_chunks.append({"type": "text", "text": ""})
                response_chunks[-1]["text"] += chunk.choices[0].delta.content

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, list), "Response should be a list"
    weather_tool = next(
        (chunk for chunk in response_chunks 
        if chunk.get("type") == "tool_use" and chunk.get("name") == "get_weather"),
        None
    )
    assert weather_tool is not None, "Weather tool call not found"
    assert isinstance(weather_tool["id"], str) and weather_tool["id"], "Tool call should have a non-empty ID"
    assert isinstance(weather_tool["input"], str), "Tool input should be a string"
    assert "London" in weather_tool["input"], "Tool input should contain 'London'"


@pytest.mark.asyncio
async def test_response_format():
    json_schema = {
        "title": {"type": "string"},
        "author": {"type": "string"},
        "genre": {"type": "string"},
        "publication_year": {"type": "integer"},
        "summary": {"type": "string"}
    }

    memory = {
        "role": "user",
        "content": [{"type": "text", "text": "Summarize The Eras Tour by Taylor Swift with fields for title, author, genre, publication year, and a short summary."}]
    }

    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "model": "gemini-2.0-flash-lite-preview-02-05",
        "json_schema": json_schema
    }
    model_args = model.model_validate(model_args)

    response_chunks = ""
    async for chunk in handler(**dict(model_args)):
        if isinstance(chunk, ChatCompletionChunk):
            if chunk.choices and chunk.choices[0].delta.content:
                response_chunks += chunk.choices[0].delta.content or ""  

    assert response_chunks, "Response should not be empty"
    assert isinstance(response_chunks, str), "Response should be a string"
    assert json.loads(response_chunks), "Response should match the JSON schema"


@pytest.mark.asyncio
async def test_invalid_model():
    memory = Message.validate({
        "role": "user",
        "content": [{"type": "text", "text": "Hello"}]
    })
    argspec = inspect.getfullargspec(handler)
    model = create_model_from_argspec("TestModel", argspec)
    model_args = {
        "memory": [memory],
        "system_prompt": "You are a helpful assistant.",
        "model": "gemini-invalid-model"
    }
    model_args = model.model_validate(model_args)

    with pytest.raises(openai.NotFoundError):
        async for _ in handler(**dict(model_args)):
            pass



import base64
import pathlib

import pytest
from anthropic.types import (
    TextBlock as AnthropicTextBlock,
)
from anthropic.types import (
    ToolUseBlock as AnthropicToolUseBlock,
)
from openai.types.chat import (
    ChatCompletionMessageToolCall as OpenAIToolCall,
)
from openai.types.chat.chat_completion_message_tool_call import Function as OpenAIFunction
from timbal.types import Content, File, FileContent, TextContent, ToolResultContent, ToolUseContent


def test_content_validation_with_text() -> None:
    content = Content.model_validate({"type": "text", "text": "Hello, World!"})
    assert isinstance(content, TextContent)
    assert content.text == "Hello, World!"
    assert content.type == "text"

    # text must be a string
    with pytest.raises(ValueError):
        Content.model_validate({"type": "text", "text": 123})


def test_text_to_openai_input() -> None:
    text_content = TextContent(text="Hello, World!")
    assert text_content.to_openai_input() == {"type": "text", "text": "Hello, World!"}


def test_text_to_anthropic_input() -> None:
    text_content = TextContent(text="Hello, World!")
    assert text_content.to_anthropic_input() == {"type": "text", "text": "Hello, World!"}


def test_text_from_anthropic_input() -> None:
    text_block = AnthropicTextBlock(
        type="text",
        text="Hello, World!"
    )
    text_content = Content.model_validate(text_block)
    assert isinstance(text_content, TextContent)
    assert text_content.text == "Hello, World!"
    assert text_content.type == "text"


def test_content_validation_with_file(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content)
    content = Content.model_validate({"type": "file", "file": File.validate(str(test_file))})
    assert isinstance(content, FileContent)
    assert isinstance(content.file, File)
    assert content.type == "file"

    # file must be a File
    with pytest.raises(ValueError):
        Content.model_validate({"type": "file", "file": "not a file"})


def test_file_to_openai_input(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content) 
    file = File.validate(str(test_file))
    file_content = FileContent(file=file)

    with open(test_file, "rb") as f:
        png_content = f.read()
        base64_content = base64.b64encode(png_content).decode("utf-8")
        data_url = f"data:image/png;base64,{base64_content}"

    assert file_content.to_openai_input() == {"type": "image_url", "image_url": {"url": data_url}}


def test_file_to_anthropic_input(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / "image.png"
    png_content = bytes.fromhex(
        '89504e470d0a1a0a'  # PNG signature
    )
    test_file.write_bytes(png_content) 
    file = File.validate(str(test_file))
    file_content = FileContent(file=file)
    assert file_content.to_anthropic_input() == {"type": "image", "source": {"type": "base64", "media_type": "image/png", "data": base64.b64encode(png_content).decode("utf-8")}}


def test_content_validation_with_tool_use() -> None:
    content = Content.model_validate({"type": "tool_use", "id": "123", "name": "tool_name", "input": {"city": "London"}})
    assert isinstance(content, ToolUseContent)
    assert content.id == "123"
    assert content.name == "tool_name"
    assert content.input == {"city": "London"}
    assert content.type == "tool_use"

    # input must be a dict
    with pytest.raises(ValueError):
        Content.model_validate({"type": "tool_use", "id": "123", "name": "get_weather", "input": "not a dict"})

    # type must be tool_use
    with pytest.raises(ValueError):
        Content.model_validate({"type": "tool", "id": "123", "name": "get_weather", "input": {"city": "London"}})


def test_tool_use_to_openai_input() -> None:
    tool_use_content = ToolUseContent(id="123",
            name="get_weather",
            input={"city": "London"})
    assert tool_use_content.to_openai_input() == {
            "id": "123",
            "type": "function",
            "function": {
                "arguments": '{"city": "London"}',
                "name": "get_weather"
            }
        }
    

def test_tool_use_from_openai_input() -> None:
    tool_use_call = OpenAIToolCall(
        id="call_62136354",
        type="function",
        function=OpenAIFunction(arguments='{"order_id":"order_12345"}', name='get_delivery_date')
    )
    tool_use_content = Content.model_validate(tool_use_call)
    assert isinstance(tool_use_content, ToolUseContent)

        
def test_tool_use_to_anthropic_input() -> None:
    tool_use_content = ToolUseContent(id="123", name="get_weather", input={"city": "London"})
    assert tool_use_content.to_anthropic_input() == {"type": "tool_use", "id": "123", "name": "get_weather", "input": {"city": "London"}}


def test_tool_use_from_anthropic_input() -> None:
    tool_use_block = AnthropicToolUseBlock(
        id="123",
        type="tool_use",
        name="get_weather",
        input={"city": "London"}
    )
    tool_use_content = Content.model_validate(tool_use_block)
    assert isinstance(tool_use_content, ToolUseContent)


def test_content_validation_with_tool_result() -> None:
    content = Content.model_validate({"type": "tool_result", "id": "123", "content": [TextContent(text="Hello, World!")]})
    assert isinstance(content, ToolResultContent)
    assert content.id == "123"
    assert content.content == [TextContent(text="Hello, World!")]
    assert content.content[0].text == "Hello, World!"
    assert content.type == "tool_result"

    # content must exist
    with pytest.raises(ValueError):
        Content.model_validate({"type": "tool_result", "id": "123", "content": {}})

    # type must be tool_result
    with pytest.raises(ValueError):
        Content.model_validate({"type": "tool", "id": "123", "content": [TextContent(text="Hello, World!")]})


def test_tool_result_to_openai_input() -> None:
    tool_result_content = ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])
    assert tool_result_content.to_openai_input() == {"role": "tool", "content": [{"type": "text", "text": "Hello, World!"}], "tool_call_id": "123"}


def test_tool_result_to_anthropic_input() -> None:
    tool_result_content = ToolResultContent(id="123", content=[TextContent(text="Hello, World!")])
    assert tool_result_content.to_anthropic_input() == {"type": "tool_result", "tool_use_id": "123", "content": [{"type": "text", "text": "Hello, World!"}]}



from typing import Any

from pydantic import BaseModel, ConfigDict

from .data import BaseData


class Snapshot(BaseModel):
    """Flow snapshot taken during a specific run.

    We allow for some fields to be None, so we can initialize this object without having to 
    await for the run to complete and send intermediate events with this.
    """
    model_config = ConfigDict(extra="ignore")

    v: str 
    """Versioning."""
    id: str
    """The snapshot's run identifier."""
    parent_id: str | None = None
    """The parent run's identifier. When null, the snapshot is a from a root run."""
    path: str
    """Snapshots can be stored by any Flow instance. Users can nest as many subflows as
    they want (with the same ids). We need this to uniquely identify where this snapshot comes from.
    """
    input: Any
    """Kwargs passed to Flow.run()"""
    output: Any | None = None
    """Output of Flow._collect_outputs()"""
    error: Any | None = None
    """Error raised during the execution of the flow."""
    t0: int 
    """Start time of the snapshot since epoch in ms."""
    t1: int | None = None
    """End time of the snapshot since epoch in ms."""
    steps: dict[str, Any] | None = None
    """Run data of each step at the end of the run."""
    data: dict[str, BaseData] | None = None
    """Data state of the flow at the end of the run."""



import asyncio
import copy
from collections.abc import AsyncGenerator, Generator
from typing import Any

import structlog
from anthropic.types import (
    InputJSONDelta,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    TextDelta,
    ToolUseBlock,
)
from pydantic import BaseModel, ConfigDict

from ..types.events import (
    AnthropicEvent,
    FlowOutputEvent,
    OpenAIEvent,
    StepChunkEvent,
    StepOutputEvent,
    StepStartEvent,
    TimbalEvent,
)

logger = structlog.get_logger("timbal.graph.stream")


async def sync_to_async_gen(gen: Generator[Any, None, None], loop: asyncio.AbstractEventLoop) -> AsyncGenerator[Any, None]:
    """Auxiliary function to convert a sync generator to an async generator."""
    while True:
        # StopIteration is special in Python. It's used to implement generator protocol and can't
        # be pickled/transferred across threads properly. By catching it explicitly in the executor 
        # function and converting it to a sentinel value, we avoid problematic exception propagation.
        def _next():
            try:
                return next(gen)
            except StopIteration: 
                return None
        value = await loop.run_in_executor(None, _next)
        if value is None:
            break
        yield value
    

class AsyncGenState(BaseModel):
    """Tracks state for async generators during flow execution.

    Attributes:
        gen (AsyncGenerator[Any, None]): The async generator being processed
        input (Any): The input for the step (before validation - as sent by the user)
        results (list[Any]): Accumulated results from generator yields (e.g. LLM message chunks)
        usage (dict[str, Any]): Resource usage stats (e.g. token counts for LLM calls)
    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    gen: AsyncGenerator[Any, None]
    input: Any
    results: list[Any] = []
    usage: Any | None = None


def handle_openai_event(
    openai_event: OpenAIEvent,
    async_gen_state: AsyncGenState,
) -> Any:
    """Process OpenAI streaming events and update the async generator state.

    Handles different types of OpenAI events including:
    - Usage statistics (token counts)
    - Tool calls (function calling)
    - Text content deltas

    Args:
        openai_event: The OpenAI event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Text chunk if the event contains content or tool call info, None otherwise
    """

    # OpenAI SDKs use a message with empty choices to indicate the end of a stream and send the usage.
    if not len(openai_event.choices):
        # ChatCompletionChunk(
        #     id='chatcmpl-B9UwHoIGJ9SzsqLMIGKaeFl87kyn1', 
        #     choices=[], 
        #     created=1741603581, 
        #     model='gpt-4o-mini-2024-07-18', 
        #     object='chat.completion.chunk', 
        #     service_tier='default', 
        #     system_fingerprint='fp_06737a9306', 
        #     usage=CompletionUsage(
        #         completion_tokens=10, 
        #         prompt_tokens=8, 
        #         total_tokens=18, 
        #         completion_tokens_details=CompletionTokensDetails(
        #             accepted_prediction_tokens=0, 
        #             audio_tokens=0, 
        #             reasoning_tokens=0, 
        #             rejected_prediction_tokens=0
        #         ), 
        #         prompt_tokens_details=PromptTokensDetails(
        #             audio_tokens=0, 
        #             cached_tokens=0
        #         )
        #     )
        # )
        # TODO Grab input cached tokens, audio tokens, etc.
        if openai_event.usage:
            if async_gen_state.usage is None:
                async_gen_state.usage = {}
            output_tokens = async_gen_state.usage.get("output_tokens", 0)
            output_tokens += openai_event.usage.completion_tokens
            input_tokens = async_gen_state.usage.get("input_tokens", 0)
            input_tokens += openai_event.usage.prompt_tokens
            async_gen_state.usage["output_tokens"] = output_tokens
            async_gen_state.usage["input_tokens"] = input_tokens
        return None

    if openai_event.choices[0].delta.tool_calls:
        tool_call = copy.deepcopy(openai_event.choices[0].delta.tool_calls[0])
        text_chunk = ""
        if tool_call.id:
            chunk_result = {
                "type": "tool_use",
                "id": tool_call.id,
                "name": tool_call.function.name,
                "input": ""
            }
            if len(async_gen_state.results):
                text_chunk = ".\n"
            text_chunk += f"Using tool '{tool_call.function.name}' with input: "
            async_gen_state.results.append(chunk_result)
        else:
            text_chunk += tool_call.function.arguments
            async_gen_state.results[-1]["input"] += tool_call.function.arguments
        return text_chunk

    if openai_event.choices[0].delta.content:
        text_chunk = openai_event.choices[0].delta.content
        chunk_result = {
            "type": "text",
            "text": text_chunk,
        }

        # Extra is allowed in the async gen pydantic model.
        if hasattr(openai_event, "citations"):
            async_gen_state.citations = openai_event.citations

        if len(async_gen_state.results):
            async_gen_state.results[-1]["text"] += text_chunk
        else:
            async_gen_state.results.append(chunk_result)
        return text_chunk


def handle_anthropic_event(
    anthropic_event: AnthropicEvent,
    async_gen_state: AsyncGenState,
) -> Any:
    """Process Anthropic streaming events and update the async generator state.

    Handles different types of Anthropic events including:
    - Message start/usage info
    - Content block starts (text or tool use)
    - Content deltas (text or JSON)
    - Message completion with usage stats

    Args:
        anthropic_event: The Anthropic event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Text chunk if the event contains content or tool input, None otherwise
    """

    # # The raw content block stop event is sent when the LLM sends multiple blocks.
    # # The raw message stop event is sent when the LLM is actually done.
    # if isinstance(anthropic_event, RawContentBlockStopEvent | RawMessageStopEvent):
    #     pass

    # 
    if isinstance(anthropic_event, RawMessageStartEvent):
        if anthropic_event.message.usage:
            input_tokens = anthropic_event.message.usage.input_tokens
            if async_gen_state.usage is None:
                async_gen_state.usage = {}
            async_gen_state.usage["input_tokens"] = input_tokens
        return None

    #
    if isinstance(anthropic_event, RawContentBlockStartEvent):
        if isinstance(anthropic_event.content_block, ToolUseBlock):
            chunk_result = {
                "type": "tool_use",
                "id": anthropic_event.content_block.id,
                "name": anthropic_event.content_block.name,
                "input": ""
            }
            async_gen_state.results.append(chunk_result)
        else:
            chunk_result = {
                "type": "text",
                "text": anthropic_event.content_block.text
            }
            async_gen_state.results.append(chunk_result)
        return None

    # 
    if isinstance(anthropic_event, RawContentBlockDeltaEvent):
        if isinstance(anthropic_event.delta, InputJSONDelta):
            text_chunk = anthropic_event.delta.partial_json
            async_gen_state.results[-1]["input"] += text_chunk
            return text_chunk
        elif isinstance(anthropic_event.delta, TextDelta):
            text_chunk = anthropic_event.delta.text
            async_gen_state.results[-1]["text"] += text_chunk
            return text_chunk

    # TODO Check if there is something in here (besides usage)
    if isinstance(anthropic_event, RawMessageDeltaEvent):
        if anthropic_event.usage:
            output_tokens = anthropic_event.usage.output_tokens
            if async_gen_state.usage is None:
                async_gen_state.usage = {}
            async_gen_state.usage["output_tokens"] = output_tokens
        return None


def handle_timbal_event(
    timbal_event: TimbalEvent,
    async_gen_state: AsyncGenState,
) -> dict[str, Any] | None:
    """Process Timbal flow events and update the async generator state.

    Handles different types of Timbal events including:
    - Step start events (filtered out for parent flows)
    - Step chunk events (currently filtered)
    - Step output events (captures usage statistics)
    - Flow output events (captures final results)

    Args:
        timbal_event: The Timbal event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        None - Timbal events currently don't return streaming content
    """

    # For a parent flow, we don't need to stream the subflow step start events.
    if isinstance(timbal_event, StepStartEvent):
        return None

    # TODO Think how should we stream partial flow outputs results or chunks. 
    # Perhaps we should only stream up if the step output is selected as a flow output.
    if isinstance(timbal_event, StepChunkEvent):
        return None
    
    # TODO Think how should we stream partial flow outputs results or chunks. 
    # Perhaps we should only stream up if the step output is selected as a flow output.
    if isinstance(timbal_event, StepOutputEvent):
        return None

    # TODO Here probably we should inject properties and convert to StepOutputEvent, no?
    if isinstance(timbal_event, FlowOutputEvent):
        async_gen_state.results = timbal_event.output
        return None


def handle_event(
    event: Any,
    async_gen_state: AsyncGenState,
) -> dict[str, Any] | None:
    """Route events to their appropriate handlers based on type.

    Dispatches events to specialized handlers for OpenAI, Anthropic, and Timbal events.
    Provides basic handling for string events and default fallback for other types.

    Args:
        event: The event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Processed event content if available, None otherwise
    """
    if isinstance(event, OpenAIEvent):
        return handle_openai_event(event, async_gen_state)

    if isinstance(event, AnthropicEvent):
        return handle_anthropic_event(event, async_gen_state)

    if isinstance(event, TimbalEvent):
        return handle_timbal_event(event, async_gen_state)

    if isinstance(event, str):
        async_gen_state.results += event
        return event

    # ? Implement custom behavior for other event types.

    # Default behavior for any other event type.
    async_gen_state.results.append(event)
    return event



import pytest
from timbal.graph.utils import (
    get_ancestors,
    get_sources,
    get_successors,
    is_dag,
    reverse_dag,
)


@pytest.fixture
def simple_dag():
    return {
        "A": {"B", "C"},
        "B": {"D"},
        "C": {"D"},
        "D": set()
    }


@pytest.fixture
def cyclic_graph():
    return {
        "A": {"B"},
        "B": {"C"},
        "C": {"A"}
    }


@pytest.fixture
def complex_dag():
    return {
        "A": {"B", "C", "D"},
        "B": {"E", "F"},
        "C": {"F", "G"},
        "D": {"G"},
        "E": {"H"},
        "F": {"H", "I"},
        "G": {"I"},
        "H": {"J"},
        "I": {"J"},
        "J": set()
    }


@pytest.fixture
def diamond_dag():
    return {
        "A": {"B", "C"},
        "B": {"D"},
        "C": {"D"},
        "D": {"E"},
        "E": set()
    }


def test_is_dag_with_valid_dag(simple_dag):
    assert is_dag(simple_dag) is True


def test_is_dag_with_cyclic_graph(cyclic_graph):
    assert is_dag(cyclic_graph) is False


def test_get_ancestors(simple_dag):
    reversed_dag = reverse_dag(simple_dag)
    assert get_ancestors("D", reversed_dag) == {"A", "B", "C"}
    assert get_ancestors("B", reversed_dag) == {"A"}
    assert get_ancestors("A", reversed_dag) == set()


def test_get_successors(simple_dag):
    assert get_successors("A", simple_dag) == {"B", "C", "D"}
    assert get_successors("B", simple_dag) == {"D"}
    assert get_successors("D", simple_dag) == set()


def test_get_sources(simple_dag):
    assert get_sources(simple_dag) == {"A"}


def test_get_sources_multiple_sources():
    dag = {
        "A": {"C"},
        "B": {"C"},
        "C": set()
    }
    assert get_sources(dag) == {"A", "B"}


def test_complex_paths(complex_dag):
    reversed_dag = reverse_dag(complex_dag)
    assert get_ancestors("F", reversed_dag) == {"A", "B", "C"}
    assert get_ancestors("J", reversed_dag) == {"A", "B", "C", "D", "E", "F", "G", "H", "I"}
    assert get_successors("A", complex_dag) == {"B", "C", "D", "E", "F", "G", "H", "I", "J"}
    assert get_successors("F", complex_dag) == {"H", "I", "J"}
    assert get_successors("J", complex_dag) == set()


def test_disconnected_dag():
    dag = {
        "A": {"B"},
        "B": set(),
        "C": {"D"},
        "D": set(),
        "E": set()
    }
    assert get_sources(dag) == {"A", "C", "E"}
    assert get_successors("A", dag) == {"B"}
    assert get_successors("C", dag) == {"D"}
    assert get_successors("E", dag) == set()


def test_single_node():
    dag = {"A": set()}
    assert is_dag(dag) is True
    assert get_sources(dag) == {"A"}
    assert get_successors("A", dag) == set()
    assert get_ancestors("A", dag) == set()


def test_empty_graph():
    dag = {}
    assert is_dag(dag) is True
    assert get_sources(dag) == set()


def test_diamond_paths(diamond_dag):
    reversed_dag = reverse_dag(diamond_dag)
    assert get_successors("A", diamond_dag) == {"B", "C", "D", "E"}
    assert get_successors("B", diamond_dag) == {"D", "E"}
    assert get_ancestors("D", reversed_dag) == {"A", "B", "C"}
    assert get_ancestors("E", reversed_dag) == {"A", "B", "C", "D"}


def test_invalid_node():
    dag = {"A": {"B"}, "B": set()}
    with pytest.raises(KeyError):
        get_successors("C", dag)
    with pytest.raises(KeyError):
        get_ancestors("C", dag)


def test_complex_cycles():
    # Self-cycle
    assert is_dag({"A": {"A"}}) is False
    
    # Complex cycle
    complex_cycle = {
        "A": {"B"},
        "B": {"C", "D"},
        "C": {"E"},
        "D": {"F"},
        "E": {"G"},
        "F": {"G"},
        "G": {"B"}  # Creates cycle B -> C/D -> ... -> G -> B
    }
    assert is_dag(complex_cycle) is False



"""Data handling system for Timbal.

This module provides a type-safe way to handle data flow between steps. It supports
three main types of data:

Usage:

1. Direct values:
   >>> data = TypeAdapter(Data).validate_python({
   ...     "type": "value",
   ...     "value": "test"
   ... })
   >>> assert data.resolve() == "test"

2. Error values:
   >>> data = TypeAdapter(Data).validate_python({
   ...     "type": "error",
   ...     "error": "error message"
   ... })
   >>> assert data.resolve() == "error message"

3. Mapped values (referencing other steps):
   >>> data = TypeAdapter(Data).validate_python({
   ...     "type": "map",
   ...     "ancestor_id": "step_1",
   ...     "ancestor_key": "return",
   ... })
   >>> context = {
   ...     "step_1.return": TypeAdapter(Data).validate_python({
   ...         "type": "value",
   ...         "value": "test"
   ...     })
   ... }
   >>> assert data.resolve(context_data=context) == "test"

All data types are validated using Pydantic and can be created from Python dictionaries
using TypeAdapter(Data).validate_python(). The resolve() method is used to get the
final value, with mapped data requiring context from other steps.
"""

import re
from abc import ABC, abstractmethod
from typing import (
    Annotated,
    Any,
    Literal,
)

import structlog
from pydantic import BaseModel, ConfigDict, Discriminator

from ..errors import DataKeyError

DATA_KEY_PATTERN = re.compile(r"^[a-zA-Z0-9_.]{0,128}$")
DATA_KEY_INTERPOLATION_PATTERN = re.compile(r"{{([a-zA-Z0-9_.]{0,128})}}")

logger = structlog.get_logger("timbal.state.data")


class BaseData(BaseModel, ABC):
    """Abstract base class for all data types in Timbal.

    It uses Pydantic for validation and supports extra fields and assignment validation.
    """
    # Allow storing extra fields in the model.
    # Validate dynamic assignments to the model.
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    type: str
    """The type identifier for the data class."""

    @abstractmethod 
    def resolve(self, context_data: dict[str, "BaseData"] | None = None) -> Any:
        pass


def get_data_key(
    data: dict[str, BaseData], 
    data_key: str,
    default: Any | None = "__NO_DEFAULT__", # We don't use an object() sentinel value here to avoid issues with serialization.
) -> Any:
    """Retrieves the value associated with a given data key from the flow's data dictionary during execution.

    This method supports nested keys using dot notation and allows indexing into lists. 

    Args:
        data: The data dictionary containing all flow execution data.
        data_key: The key to look up in the data dictionary, which can include 
                  nested keys separated by dots and list indices.
        default: The default value to return if the key is not found.
                 If not provided, a DataKeyError will be raised.

    Returns:
        Any: The resolved value associated with the specified data key.
    """
    if not DATA_KEY_PATTERN.match(data_key):
        raise ValueError(f"Invalid data key: {data_key}. Expected a string of 0-128 alphanumeric characters, underscores, or dots.")

    data_key_parts = data_key.split(".")
    data_key_builder = ""

    data_key_found = False
    data_value = None
    for i, data_key_part in enumerate(data_key_parts): # noqa: B007
        if len(data_key_builder):
            data_key_builder = f"{data_key_builder}.{data_key_part}"
        else:
            data_key_builder = data_key_part

        if data_key_builder in data:
            data_instance = data[data_key_builder]
            assert isinstance(data_instance, BaseData), "Data value should be an instance of BaseData."
            data_value = data_instance.resolve(context_data=data)
            data_key_found = True
            break
    
    if not data_key_found:
        if default == "__NO_DEFAULT__":
            raise DataKeyError(f"Cannot find data key '{data_key}'.")
        else:
            return default
    
    if data_value is None and default != "__NO_DEFAULT__":
        return default

    # When a data key is found, we need to check if we need to grab nested items or properties.
    if len(data_key_parts) > i + 1:
        remaining_key_parts = data_key_parts[i+1:]
        for remaining_key_part in remaining_key_parts:

            # Allowing indexing into lists.
            if remaining_key_part.isdigit():
                remaining_key_part = int(remaining_key_part)
                if not isinstance(data_value, list) or len(data_value) <= remaining_key_part:
                    raise KeyError(f"Cannot get {remaining_key_part} from {data_key_builder}: {data_value}.")
                data_key_builder = f"{data_key_builder}.{remaining_key_part}"
                data_value = data_value[remaining_key_part]
                continue

            # Else we assume we're working with a string key.

            elif isinstance(data_value, dict) and remaining_key_part in data_value:
                data_key_builder = f"{data_key_builder}.{remaining_key_part}"
                data_value = data_value[remaining_key_part]
                continue

            elif hasattr(data_value, remaining_key_part):
                data_key_builder = f"{data_key_builder}.{remaining_key_part}"
                data_value = getattr(data_value, remaining_key_part)
                continue

            raise KeyError(f"Cannot get data key '{data_key}'. " \
                           f"Cannot get '{remaining_key_part}' from '{data_key_builder}': {data_value}.")

    return data_value


class DataValue(BaseData):
    """Represents a direct value in the data system.

    This class handles direct values and supports string interpolation using {{key}} syntax
    when the value is a string.
    """

    type: Literal["value"] = "value"
    value: Any
    """The actual value to be stored."""

    def resolve(self, context_data: dict[str, BaseData] | None = None) -> Any:
        data_value = self.value

        # TODO Should we perform interpolation here for nested structures, e.g. list[str] ? 
        if isinstance(self.value, str):
            replacement_keys = DATA_KEY_INTERPOLATION_PATTERN.findall(self.value)
            if replacement_keys:
                assert isinstance(context_data, dict), \
                    f"Cannot resolve data key '{replacement_keys}' in string '{self.value}', " \
                    f"because context_data is not defined."
                for replacement_key in replacement_keys:
                    replacement_value = get_data_key(context_data, replacement_key)
                    # TODO If this is an LLM message, we need to dump the pydantic model.
                    # Directly add the value as a string, no need to care for python evals.
                    data_value = data_value.replace(f"{{{{{replacement_key}}}}}", str(replacement_value))

        return data_value


class DataError(BaseData):
    """Represents an error value in the data system.

    This class is used to store and propagate error messages or error states.
    """

    type: Literal["error"] = "error"
    error: Any
    """The error value or message to be stored."""

    def resolve(self, context_data: dict[str, BaseData] | None = None) -> Any: # noqa: ARG002
        # We return the error as is, because we're matching on isisntance of the class.
        return self


class DataMap(BaseData):
    """Represents a reference to another data value in the system.

    This class allows referencing values from other steps using a key path.
    The key can use dot notation to access nested values.
    """

    type: Literal["map"] = "map"
    key: str
    """The key path to the referenced data."""
    default: Any | None = "__NO_DEFAULT__"
    """The default value to return if the key is not found."""

    def resolve(self, context_data: dict[str, BaseData] | None = None) -> Any:
        assert isinstance(context_data, dict), \
            f"Cannot resolve data key '{self.key}' because context_data is not defined."
        return get_data_key(context_data, self.key, default=self.default)


Data = Annotated[
    DataValue | DataError | DataMap,
    Discriminator("type"),
]



import os
from typing import Any, Literal

import structlog
from dotenv import load_dotenv
from openai import AsyncOpenAI

from ...types import Field, Message, Tool

logger = structlog.get_logger("timbal.steps.llms.gemini_llm")

load_dotenv()

async def handler(
    # This is not actually used inside the function. It's a hack to have it as an argument
    # so we can use it when mapping data in the flow.
    prompt: Message = Field(default=None, description="Message to send to the LLM."), # noqa: ARG001
    system_prompt: str = Field(default=None, description="System prompt to guide the LLM's behavior and role."),
    memory: list[Message] = Field(description="A list containing the conversation history between the user and the LLM"),
    model: Literal["gemini-2.0-flash", "gemini-2.0-flash-lite-preview-02-05", "gemini-1.5-flash", "gemini-1.5-flash-8b", "text-embedding-004"] | str = Field(default="gemini-2.0-flash-lite-preview-02-05", description="Name of the LLM model to use"),
    tools: list[Tool | dict] = Field(default=None, description="List of tools/functions that the LLM can call"),
    tool_choice: dict[str, Any] | str = Field(default = {"type": "auto"}, description="How the model should use the provided tools"),
    max_tokens: int = Field(default=None, description="The maximum number of tokens in the response"),
    presence_penalty: float = Field(default=0, description="Penalizes token presence on a scale of -2.0 to 2.0. Higher values encourage new topics"),
    seed: int = Field(default=None, description="Seed for deterministic sampling (Beta feature)"),
    stop: str | list[str] = Field(default=None, description="Up to 4 sequences where the model will stop generating further tokens."),
    temperature: float = Field(default=1, description="Amount of randomness injected into the response. Ranges from 0.0 to 2.0."),
    top_p: float = Field(default=1, description="Nucleus sampling parameter. Ranges from 0.0 to 1.0."),
    json_schema: dict = Field(default=None, description="The JSON input schema that the model must output"),
) -> Any:
    """This handler manages interactions with Gemini's API, processing requests with
    appropriate parameters and returning streaming responses.

    Args:
        prompt: Hack to pass an input to the LLM.
        memory: Chat history containing user and LLM messages.
        system_prompt: Instructions for the LLM to follow.
        model: Name of the Gemini model to use.
        tools: List of tools/functions the LLM can call.
        tool_choice: How the model should use the provided tools.
        max_tokens: Maximum number of tokens in the response. Defaults to 1024.
        presence_penalty: Value between -2.0 and 2.0 to penalize token presence. Defaults to 0.
        seed: Seed for deterministic sampling (Beta feature).
        stop: Sequences where token generation should stop.
        temperature: Sampling temperature between 0 and 2. Defaults to 1.
        top_p: Nucleus sampling probability threshold. Defaults to 1.
        json_schema: JSON schema for structured output.

    Returns:
        Any: The streaming response from the Gemini API.
    """
    client = AsyncOpenAI(
        api_key=os.getenv("GEMINI_API_KEY"),
        base_url="https://generativelanguage.googleapis.com/v1beta/openai/"
    )

    # Get default values for parameters
    system_prompt = system_prompt.default if hasattr(system_prompt, 'default') else system_prompt
    model = model.default if hasattr(model, 'default') else model
    tools = tools.default if hasattr(tools, 'default') else tools
    tool_choice = tool_choice.default if hasattr(tool_choice, 'default') else tool_choice
    max_tokens = max_tokens.default if hasattr(max_tokens, 'default') else max_tokens
    presence_penalty = presence_penalty.default if hasattr(presence_penalty, 'default') else presence_penalty
    seed = seed.default if hasattr(seed, 'default') else seed
    stop = stop.default if hasattr(stop, 'default') else stop
    temperature = temperature.default if hasattr(temperature, 'default') else temperature
    top_p = top_p.default if hasattr(top_p, 'default') else top_p
    json_schema = json_schema.default if hasattr(json_schema, 'default') else json_schema

    messages = []
    if system_prompt:
        messages.append({"role": "system", "content": system_prompt})
    for msg in memory:
        openai_msg = msg.to_openai_input()
        messages.append(openai_msg)

    kwargs = {
        "model": model,
        "messages": messages,
        "temperature": temperature,
        "top_p": top_p,
        "stream": True,
    }

    # Optional parameters
    if stop:
        kwargs["stop"] = stop
    if seed:
        kwargs["seed"] = seed
    if max_tokens:
        kwargs["max_completion_tokens"] = max_tokens
    if presence_penalty:
        kwargs["presence_penalty"] = presence_penalty

    # Change parameters if provided to Gemini format
    if tools:
        logger.warning("This model does not support multiple tool calls. Only one tool can be called at a time.", model=model)
        tools_openai = []
        for tool in tools:
            if isinstance(tool, Tool):
                tools_openai.append(tool.to_openai())
            else:
                tools_openai.append(tool)
        kwargs["tools"] = tools_openai
        if isinstance(tool_choice, dict):
            if tool_choice["type"] != "function":
                tool_choice = tool_choice["type"]
        kwargs["tool_choice"] = tool_choice

    if json_schema:
        kwargs["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "name": "print_json",
                "schema": {
                    "type": "object",
                    "properties": json_schema
                }
            }
        }

    response = await client.chat.completions.create(**kwargs)

    return response